import usb.util
import time
import fractions
import array
from enum import Enum
import time
import math

class RequestType(Enum):
    SEND = 0x21
    RECEIVE = 0xa1

class Request(Enum):
    DETACH    = 0x00
    DNLOAD    = 0x01
    UPLOAD    = 0x02
    GETSTATUS = 0x03
    CLRSTATUS = 0x04
    GETSTATE  = 0x05
    ABORT     = 0x06

class Command(Enum):
    SET_ADDRESS_POINTER = 0x21
    ERASE = 0x41
    READ_UNPROTECT = 0x92

class DfuState(Enum):
    APP_IDLE                = 0x00
    APP_DETACH              = 0x01
    DFU_IDLE                = 0x02
    DFU_DOWNLOAD_SYNC       = 0x03
    DFU_DOWNLOAD_BUSY       = 0x04
    DFU_DOWNLOAD_IDLE       = 0x05
    DFU_MANIFEST_SYNC       = 0x06
    DFU_MANIFEST            = 0x07
    DFU_MANIFEST_WAIT_RESET = 0x08
    DFU_UPLOAD_IDLE         = 0x09
    DFU_ERROR               = 0x0a

class DfuStatus(Enum):
    OK                 = 0x00
    ERROR_TARGET       = 0x01
    ERROR_FILE         = 0x02
    ERROR_WRITE        = 0x03
    ERROR_ERASE        = 0x04
    ERROR_CHECK_ERASED = 0x05
    ERROR_PROG         = 0x06
    ERROR_VERIFY       = 0x07
    ERROR_ADDRESS      = 0x08
    ERROR_NOTDONE      = 0x09
    ERROR_FIRMWARE     = 0x0a
    ERROR_VENDOR       = 0x0b
    ERROR_USBR         = 0x0c
    ERROR_POR          = 0x0d
    ERROR_UNKNOWN      = 0x0e
    ERROR_STALLEDPKT   = 0x0f


SIZE_MULTIPLIERS = {' ': 1, 'K': 1024, 'M' : 1024*1024}
MAX_TRANSFER_SIZE = 2048

# Order is LSB first
def _address_to_4bytes(a):
    return [ a % 256, (a >> 8)%256, (a >> 16)%256, (a >> 24)%256 ]

def _get_dfu_functional_descriptors(interfaces):
    for intf in interfaces:
        pos = 0
        desc = intf.extra_descriptors
        while len(desc):
            current_desc = desc[:desc[0]]
            desc = desc[desc[0]:]
            if current_desc[1] == 0x21:
                yield current_desc


class DfuError(Exception):
    def __init__(self, message: str, status: DfuStatus, state: DfuState, text: str):
        super().__init__(message + " Device responded with {}, {}, \"{}\".".format(status, state, text))

class DfuDevice:
    def __init__(self, device):
        self._dev = device
        self._memory = None

    def init(self):
        cfg = self._dev[0]
        cfg.set()

        dfu_desc = list(_get_dfu_functional_descriptors(cfg.interfaces()))
        assert len(dfu_desc) == 1
        dfu_desc = dfu_desc[0]

        self._max_transfer_size = dfu_desc[5] + (dfu_desc[6] << 8)
        assert self._max_transfer_size <= MAX_TRANSFER_SIZE

        self.memories = {}
        for intf in cfg.interfaces():
            # example for intf_name:
            # '@Internal Flash  /0x08000000/04*016Kg,01*064Kg,07*128Kg'
            intf_name = usb.util.get_string(self._dev, intf.iInterface)
            if intf_name.count('/') != 2:
                raise Exception(f"invalid interface name for interface {intf.iInterface}: {intf_name}")
            label, baseaddr, layout = intf_name.split('/')
            baseaddr = int(baseaddr, 0) # convert hex to decimal
            addr = baseaddr

            memory = {
                'intf': intf,
                'alt': intf.bAlternateSetting,
                'sectors': []
            }

            for sector in layout.split(','):
                repeat, size = map(int, sector[:-2].split('*'))
                size *= SIZE_MULTIPLIERS[sector[-2].upper()]
                mode = sector[-1]

                while repeat > 0:
                    # TODO: verify if the section is writable
                    memory['sectors'].append({
                        'addr': addr,
                        'len': size,
                        'mode': mode
                    })

                    addr += size
                    repeat -= 1

            name = label.rstrip().lstrip('@')
            self.memories[name] = memory

    def select_memory(self, name):
        self._memory = self.memories[name]
        self._memory['intf'].set_altsetting()

    def _control_msg(self, requestType: RequestType, request: Request, value: int, buffer):
        if self._memory is None:
            raise Exception("no memory selected")
        return self._dev.ctrl_transfer(requestType.value, request.value, value, self._memory['intf'].bInterfaceNumber, buffer, timeout=60000)

    def _detach(self, timeout):
        """
        timeout: Timeout in [??]
        """
        return self._control_msg(RequestType.SEND, Request.DETACH, timeout, None)
    
    def _dnload(self, blockNum, data):
        """
        The device's current state must be dfuIDLE or dfuDNLOAD-IDLE state for
        this command to work (see STM AN3156 Fig 5).

        blockNum: block to write to
        data: byte-array-like data to write
        Returns: number of bytes that were written.
        """
        cnt = self._control_msg(RequestType.SEND, Request.DNLOAD, blockNum, list(data))
        return cnt
    
    def _upload(self, blockNum, size):
        """
        The device's current state must be dfuIDLE or dfuUPLOAD-IDLE state for
        this command to work (see STM AN3156 Fig 3).

        blockNum: block to read from
        size: number of bytes to read
        Returns: byte array containing the data that was read
        """
        return self._control_msg(RequestType.RECEIVE, Request.UPLOAD, blockNum, size)

    def _get_status(self):
        """
        Returns: A tuple of the form (status: DfuStatus, state: DfuState, text: str)
        """
        msg = self._control_msg(RequestType.RECEIVE, Request.GETSTATUS, 0, 6)

        status = DfuStatus(msg[0])
        state = DfuState(msg[4])
        poll_timeout_ms = msg[1] + (msg[2] << 8) + (msg[3] << 16)
        text = usb.util.get_string(self._dev, msg[5])

        if poll_timeout_ms > 10000:
            raise Exception("Device requested an unreasonable timeout: {} ms".format(poll_timeout_ms))

        time.sleep(float(poll_timeout_ms) / 1000.0)
        return status, state, text
    
    def clear_status(self):
        if self._get_state() == DfuState.DFU_ERROR:
            self._control_msg(RequestType.SEND, Request.CLRSTATUS, 0, None)

    def _get_state(self):
        msg = self._control_msg(RequestType.RECEIVE, Request.GETSTATE, 0, 1)
        return DfuState(msg[0])

    def _abort(self):
        self._control_msg(RequestType.SEND, Request.ABORT, 0, None)

    def _read(self, block, size):
        return self._upload(block + 2, size)

    def _write(self, block, data):
        return self._dnload(block + 2, data)

    def _set_address(self, addr):
        self._dnload(0x0, [Command.SET_ADDRESS_POINTER.value] + _address_to_4bytes(addr))
        self._expect_state(
            [DfuState.DFU_DOWNLOAD_BUSY], [DfuState.DFU_DOWNLOAD_IDLE],
            "Failed to set address 0x{:08x}.".format(addr)
        )

    def unprotect(self):
        self._dnload(0x0, [Command.READ_UNPROTECT])
        self._expect_state(
            [DfuState.DFU_DOWNLOAD_BUSY], [DfuState.DFU_DOWNLOAD_IDLE],
            "Failed to unprotect."
        )

    def _erase(self, addr):
        self._dnload(0x0, [Command.ERASE.value] + _address_to_4bytes(addr))
        self._expect_state(
            [DfuState.DFU_DOWNLOAD_BUSY], [DfuState.DFU_DOWNLOAD_IDLE],
            "Failed to erase sector at 0x{:08x}.".format(addr)
        )

    def _leave(self):
        return self._dnload(0x0, []) # Just send an empty data.

    def _expect_state(self, busy_states, target_states, error_text):
        while True:
            status, state, text = self._get_status()
            if state in target_states:
                return
            if state not in busy_states:
                raise DfuError(error_text, status, state, text)

    def erase_sector(self, sector):
        self._expect_state([], [DfuState.DFU_IDLE, DfuState.DFU_DOWNLOAD_IDLE], "Cannot erase sector")
        self._erase(sector['addr'])

    def write_sector(self, sector, data):
        self._expect_state([], [DfuState.DFU_IDLE, DfuState.DFU_DOWNLOAD_IDLE], "Cannot write sector")

        status, state, text = self._get_status()
        if state not in [DfuState.DFU_IDLE, DfuState.DFU_DOWNLOAD_IDLE]:
            raise DfuError("Cannot write sector.", status, state, text)

        self._set_address(sector['addr'])

        transfer_size = math.gcd(sector['len'], self._max_transfer_size)
        
        for blocknum in range(int(sector['len'] / transfer_size)):
            block = data[(blocknum * transfer_size):((blocknum + 1) * transfer_size)]
            n_written = self._write(blocknum, block)
            assert n_written == len(block), n_written
            self._expect_state(
                [DfuState.DFU_DOWNLOAD_BUSY], [DfuState.DFU_DOWNLOAD_IDLE],
                "Failed to write sector at 0x{:08x}, block {}".format(sector['addr'], blocknum)
            )

    def read_sector(self, sector):
        """
        Reads data from the specified sector
        Returns: a byte array containing the data
        """
        self._expect_state([], [DfuState.DFU_IDLE, DfuState.DFU_DOWNLOAD_IDLE], "Cannot read sector.")
        self._set_address(sector['addr'])

        self._abort() # exit DNLOAD_IDLE state

        transfer_size = math.gcd(sector['len'], self._max_transfer_size)

        # Device's current state must be dfuIDLE or dfuUPLOAD-IDLE for the subsequent read() commands to work.
        
        data = array.array(u'B')
        for blocknum in range(int(sector['len'] / transfer_size)):
            device_block = self._read(blocknum, transfer_size)
            assert len(device_block) == transfer_size, len(device_block)
            data.extend(device_block)

        self._abort() # take device into DFU_IDLE
        return data

    def jump_to_application(self, address):
        self._set_address(address)
        self._leave()

        try:
            self._expect_state(
                [DfuState.DFU_MANIFEST_SYNC], [DfuState.DFU_MANIFEST],
                "Failed to exit DFU mode."
            )
            # on ODrive v3.x this passes
            # on ODrive v4.x it throws a usb.core.USBError
        except usb.core.USBError:
            pass # expected




from . import get_serial_number_str_sync
import json
import os
import tempfile
import fibre.libfibre
from .utils import OperationAbortedException, yes_no_prompt

def _property_paths(prefix, obj):
    for k in dir(obj):
        v = getattr(obj, k)
        if k.startswith('_') and k.endswith('_property'):
            yield '.'.join(prefix + [k[1:-9]]), v
        elif not k.startswith('_') and isinstance(v, fibre.libfibre.RemoteObject):
            yield from _property_paths(prefix + [k], v)

def _get_dict(obj, is_config_object, prop_dict):
    result = {}

    for k in dir(obj):
        v = getattr(obj, k)
        if k.startswith('_') and k.endswith('_property') and is_config_object:
            v = v.read()
            if isinstance(v, fibre.libfibre.RemoteObject):
                v = prop_dict[v]
                print("path:", v)
            result[k[1:-9]] = v
        elif not k.startswith('_') and isinstance(v, fibre.libfibre.RemoteObject):
            sub_dict = _get_dict(v, (k == 'config') or is_config_object, prop_dict)
            if sub_dict != {}:
                result[k] = sub_dict

    return result

def _set_dict(obj, path, config_dict, prop_dict):
    errors = []
    for (k,v) in config_dict.items():
        name = ".".join(path + [k])
        if not k in dir(obj):
            errors.append("Could not restore {}: property not found on device".format(name))
            continue
        if isinstance(v, dict):
            errors += _set_dict(getattr(obj, k), path + [k], v, prop_dict)
        else:
            try:
                remote_attribute = getattr(obj, '_' + k + '_property')
                if isinstance(v, str) and hasattr(type(remote_attribute), 'exchange') and type(remote_attribute).exchange._inputs[1][1] == 'object_ref':
                    v = prop_dict[v]
                remote_attribute.exchange(v)
            except Exception as ex:
                errors.append("Could not restore {}: {}".format(name, str(ex)))
    return errors

def get_temp_config_filename(device):
    serial_number = get_serial_number_str_sync(device)
    safe_serial_number = ''.join(filter(str.isalnum, serial_number))
    return os.path.join(tempfile.gettempdir(), 'odrive-config-{}.json'.format(safe_serial_number))

def backup_config(device, filename, logger):
    """
    Exports the configuration of an MDrive to a JSON file.
    If no file name is provided, the file is placed into a
    temporary directory.
    """

    if filename is None:
        filename = get_temp_config_filename(device)

    logger.info("Saving configuration to {}...".format(filename))

    if os.path.exists(filename):
        if not yes_no_prompt("The file {} already exists. Do you want to override it?".format(filename), True):
            raise OperationAbortedException()

    prop_dict = list(_property_paths([], device))
    print([k for k, v in prop_dict])
    data = _get_dict(device, False, {v: k for k, v in prop_dict})
    with open(filename, 'w') as file:
        json.dump(data, file)
    logger.info("Configuration saved.")

def restore_config(device, filename, logger):
    """
    Restores the configuration stored in a file 
    """

    if filename is None:
        filename = get_temp_config_filename(device)

    with open(filename) as file:
        data = json.load(file)

    logger.info("Restoring configuration from {}...".format(filename))
    prop_dict = list(_property_paths([], device))
    errors = _set_dict(device, [], data, {k: v for k, v in prop_dict})

    for error in errors:
        logger.info(error)
    if errors:
        logger.warn("Some of the configuration could not be restored.")
    
    try:
        device.save_configuration()
    except fibre.libfibre.ObjectLostError:
        pass # Saving configuration makes the device reboot
    logger.info("Configuration restored.")




import sys
import platform
import threading
import fibre

async def discovered_device(device,
                        interactive_variables, objects, discovered_devices,
                        mount, shutdown_token, logger):
    """
    Handles the discovery of new devices by displaying a
    message and making the device available to the interactive
    console
    """
    mount_result = await mount(device)
    if mount_result is None:
        logger.debug("ignoring device")
        return

    display_name, serial_number, var_name = mount_result

    if serial_number in discovered_devices:
        verb = "Reconnected"
        index = discovered_devices.index(serial_number)
    else:
        verb = "Connected"
        discovered_devices.append(serial_number)
        index = len(discovered_devices) - 1

    var_name = var_name + str(index)

    # Publish new device to interactive console
    interactive_variables[var_name] = device
    device.__sealed__ = False
    try:
        device.__name__ = var_name
    finally:
        device.__sealed__ = True
    objects.append(device)
    globals()[var_name] = device # Add to globals so tab complete works
    logger.notify("{} to {} as {}".format(verb, display_name, var_name))

    # Subscribe to disappearance of the device
    device._on_lost.add_done_callback(lambda x: lost_device(var_name, shutdown_token, logger))

def lost_device(interactive_name, shutdown_token, logger):
    """
    Handles the disappearance of a device by displaying
    a message.
    """
    if not shutdown_token[0]:
        logger.warn("Oh no {} disappeared".format(interactive_name))

def get_user_name(interactive_variables, obj):
    queue = [(k, v) for k, v in interactive_variables.items() if isinstance(v, fibre.libfibre.RemoteObject)]

    if not isinstance(obj, fibre.libfibre.RemoteObject):
        return None

    while len(queue):
        k, v = queue.pop(0)
        if v == obj:
            return k
        for key in dir(v.__class__):
            class_member = getattr(v.__class__, key)
            if not key.startswith('_') and isinstance(class_member, fibre.libfibre.RemoteAttribute):
                queue.append((k + "." + (key if not class_member._magic_getter else "_" + key + "_property"), class_member._get_obj(v)))

    return "anonymous_remote_object_" + str(self._obj_handle)

def launch_shell(args, mount,
                interactive_variables,
                objects,
                print_banner, print_help,
                logger):
    """
    Launches an interactive python or IPython command line
    interface.
    As devices are connected they are made available as
    "dev0", "dev1", ...
    The names of the variables can be customized by setting branding_short.
    """

    discovered_devices = []
    shutdown_token = [False]
    globals().update(interactive_variables)

    fibre.libfibre.get_user_name = lambda obj: get_user_name(interactive_variables, obj)

    # Connect to device
    with fibre.Domain(args.path) as domain:
        on_discovery = lambda dev: discovered_device(dev, interactive_variables, objects, discovered_devices, mount, shutdown_token, logger)
        discovery = domain.run_discovery(on_discovery)

        # Check if IPython is installed
        if args.no_ipython:
            use_ipython = False
        else:
            try:
                import IPython
                use_ipython = True
            except:
                print("Warning: you don't have IPython installed.")
                print("If you want to have an improved interactive console with pretty colors,")
                print("you should install IPython\n")
                use_ipython = False

        interactive_variables["help"] = lambda: print_help(args, len(discovered_devices) > 0)

        # If IPython is installed, embed IPython shell, otherwise embed regular shell
        if use_ipython:
            # Override help function # pylint: disable=W0612
            help = lambda: print_help(args, len(discovered_devices) > 0) 
            # to fix broken "%run -i script.py"
            locals()['__name__'] = globals()['__name__'] 

            if IPython.core.getipython.get_ipython() is None:
                console = IPython.terminal.embed.InteractiveShellEmbed.instance(banner1='')
            else:
                console = IPython.terminal.embed.InteractiveShellEmbed(banner1='')

            # hack to make IPython look like the regular console
            console.runcode = console.run_cell 
            interact = console

            # Catch ObjectLostError (since disconnect is not always an error)
            default_exception_hook = console._showtraceback
            def filtered_exception_hook(ex_class, ex, trace):
                if(ex_class.__module__+'.'+ex_class.__name__ != 'fibre.libfibre.ObjectLostError'):
                    default_exception_hook(ex_class,ex,trace)
                
            console._showtraceback = filtered_exception_hook
        else:
            # Enable tab complete if possible
            try:
                import readline # Works only on Unix
                readline.parse_and_bind("tab: complete")
            except:
                sudo_prefix = "" if platform.system() == "Windows" else "sudo "
                print("Warning: could not enable tab-complete. User experience will suffer.\n"
                    "Run `{}pip install readline` and then restart this script to fix this."
                    .format(sudo_prefix))

            import code
            console = code.InteractiveConsole(locals=interactive_variables)
            interact = lambda: console.interact(banner='')

            # Catch ObjectLostError (since disconnect is not alway an error)
            console.runcode("import sys")
            console.runcode("default_exception_hook = sys.excepthook")
            console.runcode("def filtered_exception_hook(ex_class, ex, trace):\n"
                            "  if ex_class.__module__ + '.' + ex_class.__name__ != 'fibre.libfibre.ObjectLostError':\n"
                            "    default_exception_hook(ex_class,ex,trace)")
            console.runcode("sys.excepthook=filtered_exception_hook")

        # Launch shell
        print_banner()
        logger._skip_bottom_line = True
        interact()

        shutdown_token[0] = True
        discovery.stop()




import enum
from typing import Iterable, Union, Optional, List, Tuple, Sequence, cast

class Style(enum.IntEnum):
    NONE = 0x00
    BOLD = 0x01
    ITALIC = 0x02
    UNDERLINE = 0x04

class Color(enum.Enum):
    DEFAULT = enum.auto() # the terminal's default color
    COLOR_MASK = enum.auto()
    RED = enum.auto()
    YELLOW = enum.auto()
    GREEN = enum.auto()
    CYAN = enum.auto()
    BLUE = enum.auto()
    MAGENTA = enum.auto()
    GRAY = enum.auto()


class RichText():
    """
    Represents a text consisting of styled/colored string segments.
    """
    segments: List[Tuple[str, Color, Color, Style]]

    def __init__(self, val: Union[str, 'RichText', Sequence[Union[str, 'RichText']]], foreground: Optional[Color] = None, background: Optional[Color] = None, style: Optional[Style] = None):
        """
        val: A string, a RichText object or a list of strings or RichText
        objects.
        style: If specified, overrides the style of all segments in `val`.
        If not specified, uncolored strings are set to the default color.
        """

        if not (isinstance(val, tuple) or isinstance(val, list)):
            val = [cast(Union[str, 'RichText'], val)]

        self.segments = []
        for segment in val:
            if isinstance(segment, str):
                self.segments.append((segment, Color.DEFAULT, Color.DEFAULT, Style.NONE))
            elif isinstance(segment, RichText):
                self.segments.extend(segment.segments)
            else:
                raise Exception("unsupported argument")
        
        # Override attributes
        for i, (seg_str, seg_foreground, seg_background, seg_style) in enumerate(self.segments):
            if not style is None:
                seg_style = style
            if not foreground is None:
                seg_foreground = foreground
            if not background is None:
                seg_background = background
            self.segments[i] = (seg_str, seg_foreground, seg_background, seg_style)

        # Remove empty segments
        i = 0
        while i < len(self.segments):
            if self.segments[i][0] == '':
                self.segments.pop(i)
            else:
                i += 1

        # Coalesce segments with identical attributes
        i = 1
        while i < len(self.segments):
            if self.segments[i - 1][1:] == self.segments[i][1:]:
                self.segments[i - 1] = (self.segments[i - 1][0] + self.segments.pop(i)[0],) + self.segments[i - 1][1:]
            else:
                i += 1

    def __add__(self, other):
        if isinstance(other, str) or isinstance(other, RichText):
            return RichText([self, other])
        else:
            raise NotImplementedError(f"can only concatenate str and RichText (not {type(other).__name__}) to RichText.")

    def __radd__(self, other):
        if isinstance(other, str) or isinstance(other, RichText):
            return RichText([other, self])
        else:
            raise NotImplementedError(f"can only concatenate RichText to str and RichText (not {type(other).__name__}).")
    
    def __repr__(self):
        return ''.join([s[0] for s in self.segments])

    def join(self, others: Iterable[Union['RichText', str]]):
        """
        Analogous to str.join()
        """
        segments: List[Union[str, RichText]] = []
        for item in others:
            segments.append(self)
            segments.append(item)
        return RichText(segments[1:]) if len(segments) > 0 else RichText("")

    def replace(self, old: str, new: str):
        new_segments = [
            RichText(seg[0].replace(old, new), *seg[1:])
            for seg in self.segments
        ]
        return RichText(new_segments)

_VT100Colors = {
    Color.DEFAULT: 39,
    Color.GREEN: 92,
    Color.CYAN: 96,
    Color.YELLOW: 93,
    Color.RED: 91,
    Color.GRAY: 97,
}

def print_rich_text(text: RichText):
    """
    Prints `text` to stdout using VT100 escape codes for styling.
    """
    assert isinstance(text, RichText), type(text)
    output = ''
    for substr, foreground, background, style in text.segments:
        output += '\x1b[0' # reset all text attributes
        output +=  ';' + str(_VT100Colors[foreground])
        output +=  ';' + str(_VT100Colors[background] + 10)
        if style & Style.BOLD:
            output += ';1'
        if style & Style.UNDERLINE:
            output += ';4'
        output += 'm' + substr

    if len(text.segments) > 0:
        output += '\x1b[0m' # reset all attributes
    print(output)



from . import enums, utils
'\nFunctions to support legacy firmware\n'
from .rich_text import RichText, Color, Style

def format_errors(mdrv, clear=False):
    """
    Returns a summary of the error status of the device as RichText.
    """
    lines = []

    STYLE_GOOD = (Color.GREEN, Color.DEFAULT, Style.BOLD)
    STYLE_WARN = (Color.YELLOW, Color.DEFAULT, Style.BOLD)
    STYLE_BAD = (Color.RED, Color.DEFAULT, Style.BOLD)

    axes = [(name, getattr(mdrv, name)) for name in dir(mdrv) if name.startswith('axis')]
    axes.sort()

    def decode_flags(val, enum_type):
        errorcodes = {v.value: f"{enum_type.__name__}.{v.name}" for v in enum_type}
        if val == 0:
            return [RichText("no error", *STYLE_GOOD)]
        else:
            return [RichText("Error(s):", *STYLE_BAD)] + [
                RichText(errorcodes.get((1 << bit), 'UNKNOWN ERROR: 0x{:08X}'.format(1 << bit)), *STYLE_BAD)
                for bit in range(64) if val & (1 << bit) != 0]

    def decode_enum(val, enum_type):
        errorcodes = {v.value: f"{enum_type.__name__}.{v.name}" for v in enum_type}
        return [RichText(errorcodes.get(val, 'Unknown value: ' + str(val)), *(STYLE_GOOD if (val == 0) else STYLE_BAD))]

    def decode_drv_fault(axis, val):
        if val == 0:
            return [RichText("none", *STYLE_GOOD)]
        elif not hasattr(axis, '_metadata'):
            return [RichText("metadata not loaded", *STYLE_WARN)]
        else:
            return [RichText(utils.decode_drv_faults(axis._metadata['drv'], val), *STYLE_BAD)]

    def dump_item(indent, name, obj, path, decoder):
        prefix = indent + name.strip('0123456789') + ": "
        for elem in path.split('.'):
            if not hasattr(obj, elem):
                return [prefix + RichText("not found", *STYLE_WARN)]
            obj = getattr(obj, elem)

        lines = decoder(obj)
        lines = [indent + name + ": " + lines[0]] + [
            indent + "  " + line
            for line in lines[1:]
        ]
        return lines

    lines += dump_item("", "system", mdrv, 'error', lambda x: decode_flags(x, enums.LegacyMDriveError))

    for name, axis in axes:
        lines.append(name)
        lines += dump_item("  ", 'axis', axis, 'error', lambda x: decode_flags(x, enums.AxisError))
        lines += dump_item("  ", 'motor', axis, 'motor.error', lambda x: decode_flags(x, enums.MotorError))
        # lines += dump_item("  ", 'DRV fault', axis, 'last_drv_fault', lambda x: decode_drv_fault(axis, x))
        # lines += dump_item("  ", 'sensorless_estimator', axis, 'sensorless_estimator.error', lambda x: decode_flags(x, enums.SensorlessEstimatorError))
        lines += dump_item("  ", 'encoder', axis, 'encoder.error', lambda x: decode_flags(x, enums.EncoderError))
        lines += dump_item("  ", 'controller', axis, 'controller.error', lambda x: decode_flags(x, enums.ControllerError))

    if clear:
        mdrv.clear_errors()

    if clear:
        mdrv.clear_errors()

    return RichText('\n').join(lines)




from . import database
from .enums import EncoderId, AxisState, ComponentStatus, ProcedureResult, MDriveError, Rs485EncoderMode, Protocol
from .hw_version import HwVersion
from .rich_text import RichText, Color, Style
from enum import Enum
import enum
import time
from typing import NamedTuple, List, Set, Iterable, Dict, Optional
import re
import struct
import functools

class NetType(Enum):
    AB = enum.auto()
    DIGITAL = enum.auto()
    THREE_PHASE = enum.auto()
    DC = enum.auto()
    THERMISTOR = enum.auto()
    RS485 = enum.auto()
    CAN = enum.auto()
    ENC = enum.auto()

class EntityType(Enum):
    DEVICE = enum.auto()
    DEVICE_INVERTER = enum.auto()
    DEVICE_RS485_INTF = enum.auto()
    DEVICE_INC_ENC_INTF = enum.auto()
    DEVICE_ONBOARD_ENC = enum.auto()
    DEVICE_GPIO = enum.auto()
    AXIS = enum.auto()
    AXIS_MOTOR = enum.auto()
    AXIS_ENCODER = enum.auto()

class CalibrationStatus(Enum):
    OK = enum.auto()
    NEEDED = enum.auto()
    RECOMMENDED = enum.auto()
    UNKNOWN = enum.auto()

class IssueType(Enum):
    ERROR = enum.auto() # error level issues prevent the configuration from being written onto the MDrive(s)
    WARN = enum.auto() # warn level issues don't prevent the configuration from being applied but require user attention


class ConfigType():
    """
    Base class for config objects in the machine configuration. Implements
    similar functionality to typing.NamedTuple but with less restrictions.
    TODO: elaborate
    """
    def __init_subclass__(cls):
        cls._fields = [(k, t, getattr(cls, k)) for k, t in cls.__annotations__.items()]
    def __init__(self, **kwargs):
        for k, t, d in self.__class__._fields:
            if k in kwargs:
                setattr(self, k, kwargs[k])
            else:
                if isinstance(d, list):
                    d = [*d]
                setattr(self, k, d)
    def __repr__(self):
        return self.__class__.__name__ + "(" + ", ".join(k + "=" + repr(getattr(self, k)) for k, _, _ in self._fields) + ")"
    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        def eq(a, b):
            if isinstance(a, list) and isinstance(b, list):
                return len(a) == len(b) and all(eq(a[i], b[i]) for i in range(len(a)))
            return a == b
        return all(eq(getattr(self, k), getattr(other, k)) for k, _, _ in self.__class__._fields)
    def __hash__(self):
        def custom_hash(v):
            if isinstance(v, list):
                return custom_hash(tuple(elem for elem in v))
            return hash(v)
        return hash(tuple(custom_hash(getattr(self, k)) for k, _, _ in self.__class__._fields))

    def to_json(self):
        return _dump_named_tuple(self)

class EntityRef():
    def __init__(self, parent: 'EntityRef', name: str, key, entity_type: EntityType, net_type: Optional[NetType]):
        self.parent = parent
        self.name = name
        self.key = key
        self.entity_type = entity_type
        self.net_type = net_type

    def __eq__(self, other):
        return (isinstance(other, EntityRef) and
                self.parent == other.parent and
                self.entity_type == other.entity_type and
                self.key == other.key)
    def __hash__(self):
        return hash((self.parent, self.entity_type, self.key))
    def __repr__(self):
        return f'Entity[{self.name}]'

    def fullname(self):
        if self.name is None:
            return None
        else:
            parent_name = None if self.parent is None else self.parent.fullname()
            return ('' if parent_name is None else parent_name + '.') + self.name

    def get_children(self) -> Iterable['EntityRef']:
        return []


class DeviceRef(EntityRef):
    def __init__(self, machine_ref, dev_num, config: 'DeviceConfig'):
        EntityRef.__init__(self, machine_ref, f'devices[{dev_num}]', dev_num, EntityType.DEVICE, None)
        self.num = dev_num
        self.config = config

    def get_children(self):
        product_info = _db.get_product(self.config.board)
        for i in range(len(product_info['inverters'])):
            yield EntityRef(self, f'inverters[{i}]', i, EntityType.DEVICE_INVERTER, NetType.THREE_PHASE)
        for i in range(len(product_info['inc_enc'])):
            yield EntityRef(self, f'inc_enc[{i}]', i, EntityType.DEVICE_INC_ENC_INTF, NetType.AB)
        for i in range(product_info['onboard_encoders']):
            yield EntityRef(self, f'onboard_enc[{i}]', i, EntityType.DEVICE_ONBOARD_ENC, NetType.ENC)
        for i in range(len(product_info['rs485'])):
            yield EntityRef(self, f'rs485[{i}]', i, EntityType.DEVICE_RS485_INTF, NetType.RS485)
        for k, v in product_info['io'].items():
            yield EntityRef(self, f'io[{k}]', k, EntityType.DEVICE_GPIO, NetType.DIGITAL)
            if 't' in v:
                yield EntityRef(self, f'thermistor_input[{k}]', k, EntityType.DEVICE_GPIO, NetType.THERMISTOR)

class EncoderRef(EntityRef):
    def __init__(self, axis_ref, enc_num, config: 'EncoderConfig'):
        EntityRef.__init__(self, axis_ref, f'encoders[{enc_num}]', enc_num, EntityType.AXIS_ENCODER, NetType.ENC)
        self.num = enc_num
        self.config = config

    def get_children(self):
        # TODO: only yield the correct entities for this encoder
        yield EntityRef(self, f'ab', None, None, NetType.AB)
        yield EntityRef(self, f'z', None, None, NetType.DIGITAL)
        yield EntityRef(self, f'rs485', None, None, NetType.RS485)

class MotorRef(EntityRef):
    def __init__(self, axis_ref, motor_num, config: 'MotorConfig'):
        EntityRef.__init__(self, axis_ref, f'motors[{motor_num}]', motor_num, EntityType.AXIS_MOTOR, None)
        self.num = motor_num
        self.config = config

    def phases(self):
        return EntityRef(self, f'phases', None, None, NetType.THREE_PHASE)

    def get_children(self):
        yield self.phases()
        yield EntityRef(self, f'thermistor', None, None, NetType.THERMISTOR)

class AxisRef(EntityRef):
    def __init__(self, machine_ref, axis_num, config: 'AxisConfig'):
        EntityRef.__init__(self, machine_ref, f'axes[{axis_num}]', axis_num, EntityType.AXIS, None)
        self.num = axis_num
        self.config = config

    def motors(self):
        for i, config in enumerate(self.config.motors):
            yield MotorRef(self, i, config)

    def encoders(self):
        for i, config in enumerate(self.config.encoders):
            yield EncoderRef(self, i, config)

    def get_children(self):
        for motor_ref in self.motors():
            yield motor_ref
            yield from motor_ref.get_children()
        for encoder_ref in self.encoders():
            yield encoder_ref
            yield from encoder_ref.get_children()

class MachineRef(EntityRef):
    def __init__(self, config: 'MachineConfig'):
        EntityRef.__init__(self, None, None, None, None, None)
        self.config = config

    def devices(self):
        for i, dev_config in enumerate(self.config.devices):
            yield DeviceRef(self, i, dev_config)

    def axes(self):
        for i, axis_config in enumerate(self.config.axes):
            yield AxisRef(None, i, axis_config)

    def get_children(self):
        for dev_ref in self.devices():
            yield dev_ref
            yield from dev_ref.get_children()
        for axis_ref in self.axes():
            yield axis_ref
            yield from axis_ref.get_children()


_db = database.load()

_reboot_vars = [
    r'^inc_encoder[0-9]+.config.enabled$',
    r'^rs485_encoder_group0\.config\.mode$',
    r'^config\.enable_can_[a-z]$',
    r'^can\.condig\.baud_rate$',
    r'^axis[0-9]\.config\.load_encoder$',
    r'^axis[0-9]\.config\.commutation_encoder$',
    r'^axis[0-9]\.config\.can\.node_id$',
    r'^axis[0-9]\.config\.can\.is_extended$',
]

class ExpressionParseError(Exception):
    pass


def _parse_expr(val, unit, default):
    val = val.strip()
    if val == '':
        if default is None:
            return None
        val = default
    if unit != '':
        if not val.endswith(' ' + unit):
            raise ExpressionParseError(f"{val} is not a {unit}")
        val = val[:-len(unit)-1]
    try:
        return float(val)
    except ValueError:
        return ExpressionParseError(f"{val} is not a float")



class CalibrationTask():
    def __init__(self, name, func, status, issues):
        self.name = name
        self.func = func
        self.status = status
        self.issues = issues

    def run(self):
        self.func()

class IssueCollection():
    def __init__(self):
        self.issues = []

    def append(self, ref, message, level = IssueType.ERROR):
        """
        Appends an issue to the issue collection.

        `ref` defines which object in the user configuration the issue pertains
        to.
        """
        if isinstance(ref, EntityRef):
            self.issues.append((ref, message, level))
        elif isinstance(ref, list):
            for r in ref:
                self.issues.append((r, message, level))
        else:
            assert False, "bad issue reference: " + str(ref) + " - " + str(message)

    def get(self, ref: EntityRef):
        """
        Returns all issues pertaining to the specified configuration object.
        """
        for _ref, message, level in self.issues:
            if _ref == ref:
                yield message, level

    def get_for_type(self, ref_type: type):
        """
        Returns all issues pertaining to any object of the specified type.
        """
        for _ref, message, level in self.issues:
            if isinstance(_ref, ref_type):
                yield _ref, message, level

class AuxConfig(ConfigType):
    """
    Holds the configuration of a brake resistor in a user defined machine.
    """
    db_ref: str = ""
    resistance: str = ""

    @staticmethod
    def from_db(db_ref: str):
        return AuxConfig.from_json({'db_ref': db_ref})

    @staticmethod
    def from_json(json: dict) -> 'AuxConfig':
        """
        Loads a :class:`AuxConfig` object from a dictionary (usually loaded from JSON).
        """
        return _load_named_tuple(json, AuxConfig)
    
    def parse_resistance(self, unit=''): return _parse_expr(self.resistance, unit, None)

class EncoderConfig(ConfigType):
    """
    Holds the configuration of an encoder on an axis in a user defined machine.
    """
    db_ref: str = ""
    scale: str = "" # [scalar], default: 1.0

    @staticmethod
    def from_db(db_ref: str):
        return EncoderConfig.from_json({'db_ref': db_ref})

    @staticmethod
    def from_json(json: dict) -> 'EncoderConfig':
        """
        Loads a :class:`EncoderConfig` object from a dictionary (usually loaded from JSON).
        """
        return _load_named_tuple(json, EncoderConfig)

    def protocol(self):
        if self.db_ref == "":
            raise Exception("this encoder has no protocol property")
        encoder_info = _db.get_encoder(self.db_ref)
        return encoder_info['protocol']

class MotorConfig(ConfigType):
    """
    Holds the configuration of a motor on an axis in a user defined machine.
    """
    db_ref: str = ""
    scale: str = "" # [scalar], default: 1.0
    phase_resistance: str = "" # [Ohm]
    phase_inductance: str = "" # [H]
    use_thermistor: bool = False # default: False
    override: Dict = {}

    @staticmethod
    def from_db(db_ref: str, use_thermistor: bool = None, **kwargs):
        return MotorConfig.from_json({'db_ref': db_ref, 'use_thermistor': use_thermistor, **kwargs})

    @staticmethod
    def from_json(json: dict) -> 'MotorConfig':
        """
        Loads a :class:`MotorConfig` object from a dictionary (usually loaded from JSON).
        """
        return _load_named_tuple(json, MotorConfig)

    def parse_scale(self, unit=''): return _parse_expr(self.scale, unit, 1.0)
    def parse_phase_resistance(self, unit=''): return _parse_expr(self.phase_resistance, unit, None)
    def parse_phase_inductance(self, unit=''): return _parse_expr(self.phase_inductance, unit, None)

class AxisConfig(ConfigType):
    """
    Holds the configuration of an axis in a user defined machine.
    """
    motors: List[MotorConfig] = []
    encoders: List[EncoderConfig] = []
    calib_scan_vel: str = ""
    calib_scan_distance: str = ""
    calib_scan_range: str = ""
    calib_torque: str = ""
    pos_encoder: int = None
    vel_encoder: int = None
    commutation_encoder: int = None
    pos_index: str = ""
    pos_index_offset: str = ""
    can_protocol: str = ""
    can_node_id: str = ""

    @staticmethod
    def from_json(json: dict) -> 'AxisConfig':
        """
        Loads an :class:`AxisConfig` object from a dictionary (usually loaded from JSON).
        """
        return _load_named_tuple(json, AxisConfig)

    def parse_calib_scan_vel(self, unit=''): return _parse_expr(self.calib_scan_vel, unit, '0.275 rps')
    def parse_calib_scan_distance(self, unit=''): return _parse_expr(self.calib_scan_distance, unit, '1.1 turns')
    def parse_calib_scan_range(self, unit=''): return _parse_expr(self.calib_scan_range, unit, '0.02')
    def parse_calib_torque(self, unit=''): return _parse_expr(self.calib_torque, unit, None) # TODO: define default

    def add_motor(self, motor_config: MotorConfig):
        self.motors.append(motor_config)

    def add_encoder(self, encoder_type: str, cpr: int, use_for_pos: bool, use_for_vel: bool, use_for_commutation: bool):
        self.encoders.append(EncoderConfig({'type': encoder_type}))
        if use_for_pos:
            self.pos_encoder = len(self.encoders) - 1
        if use_for_vel:
            self.vel_encoder = len(self.encoders) - 1
        if use_for_commutation:
            self.commutation_encoder = len(self.encoders) - 1

class DeviceConfig(ConfigType):
    """
    Holds the configuration of an MDrive in a user defined machine.
    """
    board: HwVersion = None
    serial_number: str = ""
    shunt_conductances: List[float] = []
    aux: List[AuxConfig] = []
    fans: Dict[str, List[int]] = {}

    @staticmethod
    def from_json(json):
        return _load_named_tuple(json, DeviceConfig)

    def __init__(self, board: HwVersion, **kwargs):
        product_info = _db.get_product(board) # verify that this is a known product
        ConfigType.__init__(self, board=board, **kwargs)

JSON_FORMAT_VERSION = '0.1'

class FoundNoneError(Exception):
    pass

class FoundMultipleError(Exception):
    pass

class Net():
    @staticmethod
    def from_json(json: List[str], all_refs: Dict[str, EntityRef]):
        ports = set()
        types = set()
        for port_name in json:
            port = all_refs.get(port_name, None)
            if port is None:
                raise Exception("Unknown port: \"" + port_name + "\". Known ports are: " + str(list(all_refs.keys())))
            ports.add(port)
            types.add(port.net_type)

        assert len(types) == 1, types
        return Net(list(types)[0], ports)

    @staticmethod
    def union(net_type: NetType, *nets: Iterable['Net']):
        assert all((net.net_type == net_type for net in nets))
        if len(nets) == 0:
            return Net(net_type, set())
        return Net(net_type, set.union(*(net.ports for net in nets)))

    def __init__(self, net_type: NetType, ports: Set[EntityRef]):
        self.net_type = net_type
        self.ports = ports

    def __eq__(self, other):
        return isinstance(other, Net) and self.ports == other.ports
    def __hash__(self):
        return hash(self.ports)

    def __repr__(self):
        return 'Net(' + ', '.join(p.fullname() for p in self.ports) + ')'

    def to_json(self):
        return sorted(p.fullname() for p in self.ports)
    
    def get_all(self, entity_type: EntityType):
        """
        Returns all :class:`EntityRef`s that are connected to this net and have
        the specified type `entity_type`.
        """
        assert isinstance(entity_type, EntityType)
        def match(p):
            if p.entity_type == entity_type:
                return p
            elif p.parent is None:
                return None
            else:
                return match(p.parent)
        matches = [match(p) for p in self.ports]
        return [m for m in matches if not m is None]
    
    def get_single(self, entity_type: EntityType):
        """
        Like `get_all()`, but raises an exception if multiple or no matching
        entities were found.
        """
        ports = self.get_all(entity_type)
        if len(ports) == 0:
            raise FoundNoneError()
        elif len(ports) > 1:
            raise FoundMultipleError()
        return ports[0]

class Action():
    pass

class MotorCalibration(Action):
    def __init__(self, axis, motor_config):
        self.axis = axis
        self.motor_config = motor_config

    @property
    def name(self): return "Motor Calibration"

    @property
    def issues(self):
        if self.axis is None:
            return ["disconnected"]
        return []
    
    def run(self):
        assert len(self.issues) == 0
        time.sleep(1) # TODO: fix this
        # highly frequent (<100%) fail with current limit violation when run from test script, fixed with 1sec delay
        #  ^ does not seem to be an issue from `mtool`
        _run_state(self.axis, AxisState.MOTOR_CALIBRATION)
        print(f'measured phase_resistance: {self.axis.config.motor.phase_resistance}')
        print(f'measured phase_inductance: {self.axis.config.motor.phase_inductance}')

class EncoderCalibration(Action):
    def __init__(self, axis):
        self.axis = axis

    @property
    def name(self): return "Encoder Offset Calibration"

    @property
    def issues(self):
        if self.axis is None:
            return ["disconnected"]
        elif ComponentStatus(self.axis.commutation_mapper.status) == ComponentStatus.MISSING_INPUT:
            return ["Encoder disconnected or misconfigured."]
        return []

    def run(self):
        assert len(self.issues) == 0
        _run_state(self.axis, AxisState.ENCODER_OFFSET_CALIBRATION)

class ClosedLoopControl(Action):
    def __init__(self, axis, motor_config):
        self.axis = axis
        self.motor_config = motor_config
    
    @property
    def prerequisites(self):
        motor_calibrated = (self.motor_config.phase_resistance != "") and (self.motor_config.phase_inductance != "")
        yield MotorCalibration(self.axis, self.motor_config), CalibrationStatus.OK if motor_calibrated else CalibrationStatus.RECOMMENDED
        if ComponentStatus(self.axis.commutation_mapper.status) != ComponentStatus.NOMINAL:
            yield EncoderCalibration(self.axis), CalibrationStatus.NEEDED

class Axis():
    def __init__(self, handle, config):
        self.handle = handle
        self.config = config

    @property
    def closed_loop_control(self) -> Action:
        return ClosedLoopControl(self.handle, self.config.motors[0])


class MachineConfig(ConfigType):
    """
    Represents the configuration of a machine.

    This includes a list of axes, a list of MDrive devices and information about
    how the axes and devices are connected.
    """
    devices: List[DeviceConfig] = []
    axes: List[AxisConfig] = []
    nets: List[Net] = []

    @staticmethod
    def from_json(json: dict) -> 'MachineConfig':
        """
        Loads a :class:`MachineConfig` object from a dictionary (usually loaded from JSON).
        """
        if json.get('version', None) != JSON_FORMAT_VERSION:
            raise LoadConfigException("Unsupported config version " + str(json.get('version', None)))

        json = {**json}
        json.pop('version')
        nets = json.pop('nets') if 'nets' in json else []
        cfg = _load_named_tuple(json, MachineConfig)
        all_refs = cfg.get_entities_by_name()
        cfg.nets.extend([Net.from_json(net, all_refs) for net in nets])
        return cfg

    def __init__(self, **kwargs):
        nets = kwargs.pop('nets') if 'nets' in kwargs else []
        ConfigType.__init__(self, **kwargs)
        all_refs = self.get_entities_by_name()
        self.nets.extend([Net.from_json(net, all_refs) for net in nets])

    def to_json(self):
        json = _dump_named_tuple(self)
        json['version'] = JSON_FORMAT_VERSION
        return json

    def add_axis(self, axis_config: AxisConfig):
        """
        Adds an axis to the machine config.
        """
        self.axes.append(axis_config)
        return len(self.axes) - 1

    def add_device(self, dev_config: DeviceConfig):
        """
        Adds an MDrive device to the machine config.
        """
        self.devices.append(dev_config)
        return len(self.devices) - 1
    
    def merge(self, other: 'MachineConfig'):
        n_devices = len(self.devices)
        n_axes = len(self.axes)

        self.axes.extend(other.axes)
        self.devices.extend(other.devices)

        def shifted_name(ref: EncoderRef):
            if ref.entity_type == EntityType.DEVICE:
                my_name = f'devices[{ref.num + n_devices}]'
            elif ref.entity_type == EntityType.AXIS:
                my_name = f'axes[{ref.num + n_axes}]'
            else:
                my_name = ref.name
            parent_name = None if ref.parent is None else shifted_name(ref.parent)
            return my_name if (parent_name is None) else (parent_name + '.' + my_name)

        for net in other.nets:
            self.connect(*(shifted_name(ref) for ref in net.ports))

    def get_entities_by_name(self):
        machine_ref = MachineRef(self)
        return {ref.fullname(): ref for ref in machine_ref.get_children()}

    def connect(self, *refs):
        refs = set(refs)
        if len(refs) == 0:
            return

        all_ports = self.get_entities_by_name()
        ports = {all_ports[ref] for ref in refs}

        types = set(p.net_type for p in ports)
        if len(types) != 1:
            raise Exception(f"Cannot connect nets of different types: {types}")
        net_type = list(types)[0]
        
        for net in self.get_nets(net_type):
            intersection = net.ports & ports
            if len(intersection) > 0:
                net.ports.update(ports)
        
        self.nets.append(Net(net_type, ports))

#    def connect_phases(self, *refs):
#        assert all([isinstance(r, InverterRef) or isinstance(r, MotorRef) for r in refs])
#        self._three_phase_connections.append(refs)
#
#    def connect_abz(self, enc_ref, inc_enc_intf_ref, z_gpio_ref = None):
#        assert isinstance(enc_ref, EncoderRef)
#        assert isinstance(inc_enc_intf_ref, IncEncIntfRef)
#        assert z_gpio_ref is None or isinstance(z_gpio_ref, GpioRef)
#        # TODO: register connection for index signal
#        print("CONNECTING ", enc_ref.to_json(), inc_enc_intf_ref.to_json())
#        self._ab_connections.append([enc_ref, inc_enc_intf_ref])
#
#    def connect_rs485(self, enc_ref, rs485_intf_ref):
#        assert isinstance(enc_ref, EncoderRef)
#        assert isinstance(rs485_intf_ref, Rs485IntfRef)
#        self._rs485_connections.append([enc_ref, rs485_intf_ref])

    def get_nets(self, net_type: NetType):
        return [net for net in self.nets if net.net_type == net_type]

    def get_net(self, port: EntityRef):
        for net in self.nets:
            if any(p == port for p in net.ports):
                return net
        return Net(port.net_type, set())

    def get_status(self, mdrives):
        """
        Returns various information about the machine configuration taking into
        account the list of currently connected MDrives and their state.

        Returns a tuple (mdrv_list, output_configs, issues, axis_obj)
        where:

        mdrv_list: A list of MDrive objects that need to be configured. Each
            entry corresponds to one device in this configuration object. Some
            entries can be None.
        output_configs: A list of multi-level dictionaries that hold all
            configuration settings for all MDrives, whether they are connected or
            not. The order an length of this list corresponds to `mdrv_list`.
        issues: An `IssueCollection` containing all errors and warnings that
            were found.
        needs_reboot: A list of booleans indicating for each device if a reboot
            is required after applying the new configuration.
        axis_calib: A list of lists of CalibrationTask objects representing the
            available calibration tasks for this axis.
            Each list in axis_calib corresponds to an axis in this configuration.
        """

        output_configs = [{'config': {
                **{f'inverter{i}': {}
                for i in range(len(_db.get_product(d.board)['inverters']))},
                **{f'brake_resistor{i}': {}
                for i in range(len(_db.get_product(d.board)['aux_inverters']))}
            },
            **{
                f'axis{i}': {
                    'controller': {'config': {}}, 'config': {'motor': {}, 'can': {}},
                    'motor': {}
                    }
                for i in range(len(_db.get_product(d.board)['inverters']))
            }}
            for d in self.devices]
        issues = IssueCollection()
        axis_obj = [[] for _ in range(len(self.axes))]
        encoder_ids = [{} for _ in range(len(self.devices))]

        ref = MachineRef(self)

        # Associate devices in the configuration with connected devices
        mdrv_list = [None for _ in range(len(self.devices))]
        mdrives_by_serial_number = {mdrv._serial_number: mdrv for mdrv in mdrives}
        for dev_ref in ref.devices():
            dev_config = dev_ref.config
            if dev_config.serial_number is None:
                issues.append(dev_ref, 'Not associated with any serial number.')
            elif not dev_config.serial_number in mdrives_by_serial_number:
                issues.append(dev_ref, 'Not connected.')
            else:
                dev = mdrives_by_serial_number[dev_config.serial_number]
                board = dev._board
                if dev_config.board != board:
                    issues.append(dev_ref, 'Expected {} but found {}.'.format(dev_config.board, board))
                else:
                    mdrv_list[dev_ref.num] = dev

        for mdrv in mdrives:
            if not mdrv._serial_number in [c.serial_number for c in self.devices]:
                issues.append(ref, "Unused MDrive: " + mdrv._serial_number, IssueType.WARN)

        # Configure RS485 encoders
        for net in self.get_nets(NetType.RS485):
            try:
                rs485_intf_ref = net.get_single(EntityType.DEVICE_RS485_INTF)
            except FoundNoneError:
                continue # not an error
            except FoundMultipleError as ex:
                issues.append(ex.refs, "Multiple MDrives are not allowed on the same RS485 bus.")
                continue

            enc_refs = net.get_all(EntityType.AXIS_ENCODER)
            mdrv_output_config = output_configs[rs485_intf_ref.parent.num]

            for enc_ref in enc_refs:
                encoder_config = enc_ref.config
                enc_data = _db.get_encoder(encoder_config.db_ref)

                rs485_protocol = encoder_config.protocol()
                if not hasattr(Rs485EncoderMode, rs485_protocol):
                    issues.append(enc_ref, f"RS485 protocol {rs485_protocol} not supported.")
                    continue # ignore encoder

                # Each rs485_encoder_group is dedictated to one RS485 port on the MDrive.
                # One port can talk to multiple encoders (needs firmware change!).
                if f'rs485_encoder_group{rs485_intf_ref.key}' in mdrv_output_config:
                    rs485_encoder_group_config = mdrv_output_config[f'rs485_encoder_group{rs485_intf_ref.key}']
                else:
                    
                    rs485_encoder_group_config = {'config': {'mode': getattr(Rs485EncoderMode, rs485_protocol)}}
                    mdrv_output_config[f'rs485_encoder_group{rs485_intf_ref.key}'] = rs485_encoder_group_config
                    
                encoder_ids[rs485_intf_ref.parent.num][enc_ref] = EncoderId.RS485_ENCODER0

        # Configure incremental encoders
        for net in self.get_nets(NetType.AB):
            try:
                enc_ref = net.get_single(EntityType.AXIS_ENCODER)
            except FoundNoneError:
                continue # not an error
            except FoundMultipleError as ex:
                issues.append(ex.refs, "Multiple incremental encoders cannot share the same A/B signals.")
                continue

            inc_enc_intf_refs = net.get_all(EntityType.DEVICE_INC_ENC_INTF)
            enc_data = _db.get_encoder(enc_ref.config.db_ref)

            for inc_enc_intf_ref in inc_enc_intf_refs:
                output_configs[inc_enc_intf_ref.parent.num]["inc_encoder{}".format(inc_enc_intf_ref.key)] = {
                    'config': {'enabled': True, 'cpr': enc_data['cpr']}
                }
                encoder_ids[inc_enc_intf_ref.parent.num][enc_ref] = [EncoderId.INC_ENCODER0, EncoderId.INC_ENCODER1][inc_enc_intf_ref.key]

        # Configure onboard encoders
        for net in self.get_nets(NetType.ENC):
            try:
                axis_enc_ref = net.get_single(EntityType.AXIS_ENCODER)
                onboard_enc_ref = net.get_single(EntityType.DEVICE_ONBOARD_ENC)
            except FoundNoneError:
                continue # not an error
            except FoundMultipleError as ex:
                issues.append(ex.refs, "Only 1:1 connections supported for onboard encoders.")
                continue
            encoder_ids[onboard_enc_ref.parent.num][axis_enc_ref] = [EncoderId.ONBOARD_ENCODER0, EncoderId.ONBOARD_ENCODER1][onboard_enc_ref.key]

        # Configure motors
        for net in self.get_nets(NetType.THREE_PHASE):
            try:
                inv_ref = net.get_single(EntityType.DEVICE_INVERTER)
            except FoundNoneError:
                continue # not connected to any inverter
            except FoundMultipleError:
                issues.append(axis_ref, "Phase bundling not implemented. Each motor must be connected to at most one inverter.")
                continue

            try:
                motor_ref = net.get_single(EntityType.AXIS_MOTOR)
            except FoundNoneError:
                continue # not connected to any inverter
            except FoundMultipleError:
                issues.append(axis_ref, "Each inverter must be connected to at most one motor.")
                continue

            motor_config = motor_ref.config
            motor_data = _db.get_motor(motor_config.db_ref)

            axis_output_config = output_configs[inv_ref.parent.key]['axis{}'.format(inv_ref.key)]

            if motor_config.scale != '':
                issues.append(motor_ref, "Support for motor scale other than 1.0 not implemented.")

            if motor_config.phase_resistance != "":
                axis_output_config['config']['motor']['phase_resistance'] = motor_config.parse_phase_resistance('Ohm')
            else:
                axis_output_config['config']['motor']['phase_resistance'] = motor_data['phase_resistance']

            if motor_config.phase_inductance != "":
                axis_output_config['config']['motor']['phase_inductance'] = motor_config.parse_phase_inductance('H')
            else:
                axis_output_config['config']['motor']['phase_inductance'] = motor_data['phase_inductance']

            if 'calibration_current' in motor_data:
                axis_output_config['config']['motor']['calibration_current'] = motor_data['calibration_current']

            # TODO: take into account user max current
            # Note: we multiply the motor current limit by two since it's given in "continuous max"
            inv_data = _db.get_product(self.devices[inv_ref.parent.key].board)['inverters'][inv_ref.key]
            axis_output_config['config']['motor']['current_soft_max'] = min(inv_data['max_current'], 2 * motor_data['max_current']) # TODO: inverter max current should be set as a separate config var
            axis_output_config['config']['motor']['current_hard_max'] = 1.5 * min(inv_data['max_current'], 2 * motor_data['max_current'])

            axis_output_config['config']['motor']['pole_pairs'] = motor_data['pole_pairs']
            axis_output_config['config']['motor']['torque_constant'] = motor_data['torque_constant']
            axis_output_config['config']['motor']['phase_resistance_valid'] = True
            axis_output_config['config']['motor']['phase_inductance_valid'] = True

        # Configure thermistors
        for net in self.get_nets(NetType.THERMISTOR):
            try:
                motor_ref = net.get_single(EntityType.AXIS_MOTOR)
            except FoundNoneError:
                continue # not connected to any inverter
            except FoundMultipleError:
                issues.append(motor_ref, "Each thermistor must be connected to at most one motor.")
                continue
            try:
                gpio_ref = net.get_single(EntityType.DEVICE_GPIO)
            except FoundNoneError:
                continue # not connected to any inverter
            except FoundMultipleError:
                issues.append(gpio_ref, "Each thermistor must be connected to at most one MDrive IO.")
                continue

            motor_config = motor_ref.config
            motor_data = {**_db.get_motor(motor_config.db_ref), **motor_config.override}
            # if motor_ref.num != gpio_ref.num:
            #     issues.append(motor_ref, f"The motor thermistor must either be disconnected or connected to the thermistor input that corresponds to the same MDrive and axis to which the motor is connected.")
            #     continue

            axis_output_config = output_configs[gpio_ref.parent.key]['axis{}'.format(0)]

            # axis_output_config = mdrv_output_config['axis{}'.format(temp_in_num)]

            mdrv_data = _db.get_product(gpio_ref.parent.config.board)

            temp_in_data = [temp_data for temp_data in  mdrv_data['temp_in'] if temp_data['io'] == gpio_ref.key][0]

            axis_output_config['motor']['motor_thermistor'] = {
                'config': {
                    'r_ref': motor_data['thermistor_r25'],
                    'beta': motor_data['thermistor_beta'],
                    'temp_limit_lower': motor_data['min_temp'] if 'min_temp' in motor_data else motor_data['max_temp'] - 20,
                    'temp_limit_upper': motor_data['max_temp'],
                    'enabled': True # requires reboot (?)
                }
            }
            # TODO: set corresponding GPIO mode to 3 (or probably should be handled by firmware)

        # Configure shunt conductance
        for dev_num, dev_config in enumerate(self.devices):
            for inv_num in range(len(_db.get_product(dev_config.board)['inverters'])):
                if len(dev_config.shunt_conductances) > inv_num:
                    output_configs[dev_num]['config']['inverter{}'.format(inv_num)]['shunt_conductance'] = dev_config.shunt_conductances[inv_num]

        # Other configuration
        for dev_num, dev_config in enumerate(self.devices):
            output_config = output_configs[dev_num]

            # TODO: set vbus voltage trip level based on power supply setting
            # TODO: set dc_max_negative_current based on power supply setting
            for inv_num, aux in enumerate(dev_config.aux):
                if inv_num > len(_db.get_product(dev_config.board)['aux_inverters']):
                    continue # TODO: log issue if trying to config invalid aux inv
                if aux.resistance != "":
                    output_config['config'][f'brake_resistor{inv_num}']['resistance'] = aux.parse_resistance('Ohm')
                else:
                    output_config['config'][f'brake_resistor{inv_num}']['resistance'] = _db.get_brakeR(aux.db_ref)['resistance']

                # output_config['config'][f'brake_resistor{inv_num}']['enable'] = True # TODO: discuss desired default behavior

            output_config['config']['dc_max_negative_current'] = -1

            # fan config
            for id, limits in dev_config.fans.items():
                output_config['config'][id]= {
                    'lower': limits[0],
                    'upper': limits[1],
                    'enabled': True
                }

        # Configure axes
        for axis_ref in ref.axes():
            axis_config = axis_ref.config
            three_phase_net = Net.union(NetType.THREE_PHASE, *(self.get_net(motor_ref.phases()) for motor_ref in axis_ref.motors()))

            try:
                inv_ref = three_phase_net.get_single(EntityType.DEVICE_INVERTER)
            except FoundNoneError:
                issues.append(axis_ref, f"Not connected to any inverter.")
                continue
            except FoundMultipleError:
                issues.append(axis_ref, f"Connected to more than one inverters.")
                continue

            axis_output_config = output_configs[inv_ref.parent.num]['axis{}'.format(inv_ref.key)]

            if not axis_config.parse_calib_scan_vel('rps') is None:
                axis_output_config['config']['calib_scan_vel'] = axis_config.parse_calib_scan_vel('rps') * motor_data['pole_pairs'] * motor_config.parse_scale()
            
            if not axis_config.parse_calib_scan_distance('turns') is None:
                axis_output_config['config']['calib_scan_distance'] = axis_config.parse_calib_scan_distance('turns') * motor_data['pole_pairs'] * motor_config.parse_scale()
            
            if not axis_config.parse_calib_scan_range() is None:
                axis_output_config['config']['calib_range'] = axis_config.parse_calib_scan_range()
            
            # TODO: check if larger than current limit
            if not axis_config.parse_calib_torque('Nm') is None:
                axis_output_config['config']['calibration_lockin'] = {}
                axis_output_config['config']['calibration_lockin']['current'] = axis_config.parse_calib_torque('Nm') / motor_data['torque_constant']      

            if axis_config.pos_encoder is None:
                # TODO: use sensorless mode
                issues.append(axis_ref, "No position encoder specified")
            else:
                enc_ref = list(axis_ref.encoders())[axis_config.pos_encoder]
                enc_id = encoder_ids[inv_ref.parent.num].get(enc_ref, None)
                if enc_id is None:
                    issues.append(axis_ref, f"Load encoder of this axis must be connected to the same mwdrive as the motor ({output_configs[inv_ref.parent.num]['serial_number']})")
                else:
                    axis_output_config['config']['load_encoder'] = enc_id

            if axis_config.commutation_encoder is None:
                # TODO: use sensorless mode
                issues.append(axis_ref, "No commutation encoder specified")
            else:
                enc_ref = list(axis_ref.encoders())[axis_config.commutation_encoder]
                enc_id = encoder_ids[inv_ref.parent.num].get(enc_ref, None)
                if enc_id is None:
                    issues.append(axis_ref, f"Commutation encoder of this axis must be connected to the same mwdrive as the motor ({output_configs[inv_rev.parent.num]['serial_number']})")
                else:
                    axis_output_config['config']['commutation_encoder'] = enc_id

            if axis_config.vel_encoder is None:
                # TODO: use sensorless mode
                issues.append(axis_ref, "No commutation encoder specified")
            elif axis_config.vel_encoder == axis_config.commutation_encoder:
                axis_output_config['controller']['config']['use_commutation_vel'] = False
            elif axis_config.vel_encoder == axis_config.pos_encoder:
                axis_output_config['controller']['config']['use_commutation_vel'] = True
            else:
                issues.append(axis_ref, "The velocity encoder must be the same as either the position encoder or the commutation encoder.")

            if axis_config.can_protocol == "":
                pass
            elif axis_config.can_protocol == "simple":
                output_configs[inv_ref.parent.num]['can'] = {"config": {"protocol": Protocol.SIMPLE}}
                axis_output_config['config']['can']['node_id'] = int(axis_config.can_node_id)
            else:
                issues.append(axis_ref, f"Unknown CAN protocol {axis_config.can_protocol}.")

            #if enc_id in [EncoderId.INC_ENCODER0, EncoderId.INC_ENCODER1]:
            axis = None if mdrv_list[inv_ref.parent.num] is None else getattr(mdrv_list[inv_ref.parent.num], 'axis{}'.format(inv_ref.key))
            axis_obj[axis_ref.num] = Axis(axis, axis_config)

        def strip_empty_fields(d):
            result = {}
            for k, v in d.items():
                if isinstance(v, dict):
                    v = strip_empty_fields(v)
                    if v == {}:
                        continue
                result[k] = v
            return result

        output_configs = [strip_empty_fields(cfg) for cfg in output_configs]

        def compare(path, obj, config):
            all_equal = True
            reboot_required = False

            for k, v in config.items():
                if isinstance(v, dict):
                    equal, sub_reboot = compare(path + [k], getattr(obj, k), v)
                    reboot_required = reboot_required or sub_reboot
                elif isinstance(v, float):
                    # TODO: this comparison is a bit fragile (shouldn't compare floats like this)
                    equal = getattr(obj, k) == struct.unpack("f", struct.pack("f", v))[0]
                elif isinstance(v, enum.Enum):
                    equal = v.value == getattr(obj, k)
                else:
                    equal = getattr(obj, k) == v

                all_equal = all_equal and equal
                if not equal:
                    name = '.'.join(path + [k])
                    #print("changed: ", name)
                    if any(re.match(r, name) for r in _reboot_vars):
                        reboot_required = True
            return all_equal, reboot_required

        needs_reboot = [False] * len(mdrv_list)
        for dev_ref in ref.devices():
            mdrv = mdrv_list[dev_ref.num]
            if not mdrv is None:
                all_equal, needs_reboot[dev_ref.num] = compare([], mdrv, output_configs[dev_ref.num])
                if not all_equal:
                    issues.append(dev_ref, "Configuration needs to be committed to MDrive", IssueType.WARN)
                needs_reboot[dev_ref.num] = needs_reboot[dev_ref.num] or mdrv.reboot_required
                if needs_reboot[dev_ref.num]:
                    issues.append(dev_ref, "Reboot required", IssueType.WARN)

        return mdrv_list, output_configs, issues, needs_reboot, axis_obj
#

    def format_status(self, mdrives) -> RichText:
        """
        Returns a status summary of the configuration in a human readable format.
        This includes warnings about any issues with the configuration, calibration
        status and more.
        """

        mdrv_list, output_configs, issues, needs_reboot, axis_obj = self.get_status(mdrives)

        lines = []

        check_sign = "\u2705"
        info_sign = "\U0001F4A1"
        warning_sign = "\u26A0\uFE0F "
        error_sign = "\u274C"
        question_sign = "  " # TODO

        sign = {
            IssueType.WARN: warning_sign,
            IssueType.ERROR: error_sign
        }

        style = {
            IssueType.WARN: (Color.YELLOW, Color.DEFAULT, Style.BOLD),
            IssueType.ERROR: (Color.RED, Color.DEFAULT, Style.BOLD)
        }

        ref = MachineRef(self)

        for message, level in issues.get(ref):
            lines.append(sign[level] + " " + RichText(message, *style[level]))

        for dev_ref in ref.devices():
            dev_config = dev_ref.config
            if not dev_config.serial_number is None:
                name = "MDrive with serial number " + str(dev_config.serial_number)
            else:
                name = "MDrive {} (unspecified serial number)".format(dev_ref.num)
            lines.append(name)
            for message, level in issues.get(dev_ref):
                lines.append("  " + sign[level] + " " + RichText(message, *style[level]))

        for axis_ref in ref.axes():
            axis_config = axis_ref.config
            lines.append("Axis " + str(axis_ref.num))
            for message, level in issues.get(axis_ref):
                lines.append("  " + sign[level] + " " + RichText(message, *style[level]))
            for motor_ref in axis_ref.motors():
                for message, level in issues.get(motor_ref):
                    lines.append("  " + sign[level] + " Motor " + str(motor_ref.num) + ": " + RichText(message, *style[level]))
            for enc_ref in axis_ref.encoders():
                for message, level in issues.get(enc_ref):
                    lines.append("  " + sign[level] + " Encoder " + str(enc_ref.num) + ": " + RichText(message, *style[level]))


            # TODO: the calibration status is not really meaningful if there is
            # uncommitted configuration. needs_reboot is not the correct thing
            # to check but it's close enough for now
            if any(needs_reboot):
                continue

            for calib, calib_status in axis_obj[axis_ref.num].closed_loop_control.prerequisites:
                if calib_status == CalibrationStatus.OK:
                    lines.append("  " + check_sign + " " + RichText(str(calib.name) + " ok", Color.GREEN))
                elif calib_status == CalibrationStatus.RECOMMENDED:
                    lines.append("  " + info_sign + " " + RichText(str(calib.name) + " recommended"))
                elif calib_status == CalibrationStatus.NEEDED:
                    lines.append("  " + warning_sign + " " + RichText(str(calib.name) + " needed", style=Style.BOLD))
                elif calib_status == CalibrationStatus.UNKNOWN:
                    lines.append("  " + question_sign + " " + RichText(str(calib.name) + ": unknown status", Color.GRAY))
                else:
                    assert(False)
                for issue in calib.issues:
                    lines.append("    " + RichText(str(calib.name) + ": " + issue, Color.RED))

        return RichText("\n").join(lines)


    def apply(self, mdrives):
        """
        Commits the configuration to the mdrives. A reboot may be needed after
        this.

        If there a are problems with the configuration this function throws an
        exception and does not change anything on any MDrive.
        In this case show_status() can be used to get more detailed error
        information.

        Returns a list of devices that need a reboot before the configuration
        takes effect.
        """

        mdrv_list, output_configs, issues, needs_reboot, axis_obj = self.get_status(mdrives)

        if any([m for _, m, level in issues.issues if level == IssueType.ERROR]):
            print([m for _, m, level in issues.issues if level == IssueType.ERROR])
            raise Exception("There are problems with this configuration. No changes were applied to the MDrive(s).")

        def _apply(obj, config):
            for k, v in config.items():
                if isinstance(v, dict):
                    _apply(getattr(obj, k), v)
                else:
                    setattr(obj, k, v)

        for mdrv_num, mdrv in enumerate(mdrv_list):
            if not mdrv is None:
                if needs_reboot[mdrv_num]:
                    mdrv.reboot_required = True
                _apply(mdrv, output_configs[mdrv_num])

        return [mdrv_list[num] for num, r in enumerate(needs_reboot) if r]

    def calibrate(self, mdrives, include_optional = True):
        """
        Runs the calibration tasks for this machine configuration based on the
        current state of the MDrives. This can include a reboot of one or more
        MDrives.
        """

        mdrv_list, output_configs, issues, needs_reboot, axis_obj = self.get_status(mdrives)

        if any(needs_reboot):
            raise Exception("Some devices need to be rebooted for the configuration to take effect.")

        for axis_num, axis in enumerate(axis_obj):
            for calib, calib_status in axis.closed_loop_control.prerequisites:
                if (calib_status == CalibrationStatus.RECOMMENDED and include_optional) or calib_status == CalibrationStatus.NEEDED:
                    issues = list(calib.issues)
                    if len(issues):
                        raise Exception(f"Can't run {calib.name}: {issues}")
                    print(f"Running {calib.name} on axis {axis_num}...")
                    calib.run()
        print("Done!")

class LoadConfigException(Exception):
    pass

def _load_py_obj(data, py_type):
    if py_type == str:
        if isinstance(data, str):
            return data
        else:
            raise LoadConfigException(f"expected str but got {data}")

    elif py_type == bool:
        if isinstance(data, bool):
            return data
        else:
            raise LoadConfigException(f"expected bool but got {data}")

    elif py_type == int:
        if isinstance(data, int):
            return data
        else:
            raise LoadConfigException(f"expected int but got {data}")

    elif (hasattr(py_type, '_name') and py_type._name == 'List') or (hasattr(py_type, '__name__') and py_type.__name__ == 'List'): # __name__ check is for Python 3.6 compatibility
        elem_type = py_type.__args__[0]
        if isinstance(data, list):
            return [_load_py_obj(v, elem_type) for v in data if not v is None]
        else:
            raise LoadConfigException(f"expected list but got {data}")

    elif hasattr(py_type, 'from_json'):
        return py_type.from_json(data)

    elif (hasattr(py_type, '_name') and py_type._name == 'Dict') or (hasattr(py_type, '__name__') and py_type.__name__ == 'Dict'): # __name__ check is for Python 3.6 compatibility
        elem_type = py_type.__args__[1]
        if isinstance(data, dict):
            return {
                i: _load_py_obj(v, elem_type) 
                for i,v in data.items() if not v is None}
        else:
            raise LoadConfigException(f"expected dict but got {data}")
        
    elif hasattr(py_type, '__name__') and py_type.__name__ in ('T', 'KT', 'VT'):
        return data
    
    else:
        # This is not the fault of the user config but of our own code
        raise Exception(f"Don't know how to decode JSON to {py_type.__name__}")

def _load_named_tuple(data, py_type, **kwargs):
    assert isinstance(data, dict)

    unsupported_keys = set(data.keys()) - (set(py_type.__annotations__.keys()) - set(kwargs.keys()))
    if len(unsupported_keys) > 0:
        raise Exception(f"Unexpected keys for {py_type.__name__}: {unsupported_keys}")

    init_dict = {}
    for field_name, field_type, field_default in py_type._fields:
        field_val = data.get(field_name, None)
        if field_val is None:
            continue
        init_dict[field_name] = _load_py_obj(field_val, field_type)

    return py_type(**init_dict, **kwargs)


def _is_default(val, default):
    if val == default:
        return True
    elif type(val) == str:
        return val.strip() == default
    return False

def _dump_py_obj(py_obj):
    if py_obj is None:
        return None
    elif type(py_obj) == str:
        return None if py_obj.strip() == "" else py_obj.strip()
    elif type(py_obj) == int:
        return py_obj
    elif type(py_obj) == float:
        return py_obj
    elif type(py_obj) == bool:
        return py_obj
    elif type(py_obj) == list:
        return None if (len(py_obj) == 0) else [_dump_py_obj(o) for o in py_obj]
    elif type(py_obj) == dict:
        return None if (len(py_obj) == 0) else {i:_dump_py_obj(o) for i,o in py_obj.items()}
    elif hasattr(py_obj, 'to_json'):
        return py_obj.to_json()
    else:
        raise Exception(f"Don't know how to dump object {py_obj}")

def _dump_named_tuple(py_obj):
    data = {}
    for k, _, d in py_obj.__class__._fields:
        subdata = getattr(py_obj, k)
        if _is_default(subdata, d):
            continue # skip
        data[k] = _dump_py_obj(subdata)
    return data


def _run_state(axis, state):
    axis.requested_state = state
    while AxisState(axis.requested_state) == state:
        time.sleep(0.1)
    while ProcedureResult(axis.procedure_result) == ProcedureResult.BUSY:
        time.sleep(0.1)

    result = ProcedureResult(axis.procedure_result)
    if result == ProcedureResult.DISARMED:
        raise Exception("Device failed with {}".format(MDriveError(axis.disarm_reason).name))
    elif result != ProcedureResult.SUCCESS:
        raise Exception("Device returned {}".format(result))



#!/bin/python

from ctypes import *
import asyncio
import os
from itertools import count, takewhile
import struct
from types import MethodType
import concurrent
import enum
import threading
import time
import platform
from .utils import Logger
import sys

# Enable this for better tracebacks in some cases
#import tracemalloc
#tracemalloc.start(10)

lib_names = {
    ('Linux', 'x86_64'): 'libfibre-linux-amd64.so',
    ('Linux', 'armv7l'): 'libfibre-linux-armhf.so',
    ('Linux', 'aarch64'): 'libfibre-linux-aarch64.so',
    ('Windows', 'AMD64'): 'libfibre-windows-amd64.dll',
    ('Darwin', 'x86_64'): 'libfibre-macos-x86.dylib',
    ('Darwin', 'arm64'): 'libfibre-macos-arm.dylib',
}

system_desc = (platform.system(), platform.machine())

if system_desc == ('Linux', 'aarch64') and sys.maxsize <= 2**32:
    system_desc = ('Linux', 'armv7l') # Python running in 32-bit mode on 64-bit OS

script_dir = os.path.dirname(os.path.realpath(__file__))
fibre_cpp_paths = [
    os.path.join(os.path.dirname(os.path.dirname(script_dir)), "cpp"),
    os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(script_dir)))), "Firmware", "fibre-cpp")
]

def get_first(lst, predicate, default):
    for item in lst:
        if predicate(item):
            return item
    return default

if not system_desc in lib_names:
    raise ModuleNotFoundError(("libfibre is not supported on your platform ({} {}). "
            "Go to https://github.com/samuelsadok/fibre-cpp/tree/devel for "
            "instructions on how to compile libfibre. Once you have compiled it, "
            "add it to this folder.").format(*system_desc))

lib_name = lib_names[system_desc]
search_paths = fibre_cpp_paths + [script_dir]

lib_path = get_first(
    (os.path.join(p, lib_name) for p in search_paths),
    os.path.isfile, None)

if lib_path is None:
    raise ModuleNotFoundError("{} was not found in {}".format(lib_name, search_paths))

if os.path.getsize(lib_path) < 1000:
    raise ModuleNotFoundError("{} is too small. Did you forget to init git lfs? Try this:\n"
        " 1. Install git lfs (https://git-lfs.github.com/)\n"
        " 2. Run `cd {}`\n"
        " 3. Run `git lfs install`\n"
        " 4. Run `git lfs pull`".format(lib_path, os.path.dirname(lib_path)))

if os.name == 'nt':
  dll_dir = os.path.dirname(lib_path)
  try:
    # New way in python 3.8+
    os.add_dll_directory(dll_dir)
  except:
    os.environ['PATH'] = dll_dir + os.pathsep + os.environ['PATH']
  lib = windll.LoadLibrary(lib_path)
else:
  lib = cdll.LoadLibrary(lib_path)

# libfibre definitions --------------------------------------------------------#

PostSignature = CFUNCTYPE(c_int, CFUNCTYPE(None, c_void_p), POINTER(c_int))
RegisterEventSignature = CFUNCTYPE(c_int, c_int, c_uint32, CFUNCTYPE(None, c_void_p, c_int), POINTER(c_int))
DeregisterEventSignature = CFUNCTYPE(c_int, c_int)
CallLaterSignature = CFUNCTYPE(c_void_p, c_float, CFUNCTYPE(None, c_void_p), POINTER(c_int))
CancelTimerSignature = CFUNCTYPE(c_int, c_void_p)

OnFoundObjectSignature = CFUNCTYPE(None, c_void_p, c_void_p, c_void_p)
OnLostObjectSignature = CFUNCTYPE(None, c_void_p, c_void_p)
OnStoppedSignature = CFUNCTYPE(None, c_void_p, c_int)

OnAttributeAddedSignature = CFUNCTYPE(None, c_void_p, c_void_p, c_void_p, c_size_t, c_void_p, c_void_p, c_size_t)
OnAttributeRemovedSignature = CFUNCTYPE(None, c_void_p, c_void_p)
OnFunctionAddedSignature = CFUNCTYPE(None, c_void_p, c_void_p, c_void_p, c_size_t, POINTER(c_char_p), POINTER(c_char_p), POINTER(c_char_p), POINTER(c_char_p))
OnFunctionRemovedSignature = CFUNCTYPE(None, c_void_p, c_void_p)

OnCallCompletedSignature = CFUNCTYPE(c_int, c_void_p, c_int, c_void_p, c_void_p, POINTER(c_void_p), POINTER(c_size_t), POINTER(c_void_p), POINTER(c_size_t))
OnTxCompletedSignature = CFUNCTYPE(None, c_void_p, c_void_p, c_int, c_void_p)
OnRxCompletedSignature = CFUNCTYPE(None, c_void_p, c_void_p, c_int, c_void_p)

kFibreOk = 0
kFibreBusy = 1
kFibreCancelled = 2
kFibreClosed = 3
kFibreInvalidArgument = 4
kFibreInternalError = 5
kFibreProtocolError = 6
kFibreHostUnreachable = 7

class LibFibreVersion(Structure):
    _fields_ = [
        ("major", c_uint16),
        ("minor", c_uint16),
        ("patch", c_uint16),
    ]

    def __repr__(self):
        return "{}.{}.{}".format(self.major, self.minor, self.patch)

class LibFibreEventLoop(Structure):
    _fields_ = [
        ("post", PostSignature),
        ("register_event", RegisterEventSignature),
        ("deregister_event", DeregisterEventSignature),
        ("call_later", CallLaterSignature),
        ("cancel_timer", CancelTimerSignature),
    ]

libfibre_get_version = lib.libfibre_get_version
libfibre_get_version.argtypes = []
libfibre_get_version.restype = POINTER(LibFibreVersion)

version = libfibre_get_version().contents
if (version.major, version.minor) != (0, 1):
    raise Exception("Incompatible libfibre version: {}".format(version))

libfibre_open = lib.libfibre_open
libfibre_open.argtypes = [LibFibreEventLoop]
libfibre_open.restype = c_void_p

libfibre_close = lib.libfibre_close
libfibre_close.argtypes = [c_void_p]
libfibre_close.restype = None

libfibre_open_domain = lib.libfibre_open_domain
libfibre_open_domain.argtypes = [c_void_p, c_char_p, c_size_t]
libfibre_open_domain.restype = c_void_p

libfibre_close_domain = lib.libfibre_close_domain
libfibre_close_domain.argtypes = [c_void_p]
libfibre_close_domain.restype = None

libfibre_start_discovery = lib.libfibre_start_discovery
libfibre_start_discovery.argtypes = [c_void_p, c_void_p, OnFoundObjectSignature, OnLostObjectSignature, OnStoppedSignature, c_void_p]
libfibre_start_discovery.restype = None

libfibre_stop_discovery = lib.libfibre_stop_discovery
libfibre_stop_discovery.argtypes = [c_void_p]
libfibre_stop_discovery.restype = None

libfibre_subscribe_to_interface = lib.libfibre_subscribe_to_interface
libfibre_subscribe_to_interface.argtypes = [c_void_p, OnAttributeAddedSignature, OnAttributeRemovedSignature, OnFunctionAddedSignature, OnFunctionRemovedSignature, c_void_p]
libfibre_subscribe_to_interface.restype = None

libfibre_get_attribute = lib.libfibre_get_attribute
libfibre_get_attribute.argtypes = [c_void_p, c_void_p, POINTER(c_void_p)]
libfibre_get_attribute.restype = c_int

libfibre_call = lib.libfibre_call
libfibre_call.argtypes = [c_void_p, POINTER(c_void_p), c_int, c_void_p, c_size_t, c_void_p, c_size_t, POINTER(c_void_p), POINTER(c_void_p), OnCallCompletedSignature, c_void_p]
libfibre_call.restype = c_int

libfibre_start_tx = lib.libfibre_start_tx
libfibre_start_tx.argtypes = [c_void_p, c_char_p, c_size_t, OnTxCompletedSignature, c_void_p]
libfibre_start_tx.restype = None

libfibre_cancel_tx = lib.libfibre_cancel_tx
libfibre_cancel_tx.argtypes = [c_void_p]
libfibre_cancel_tx.restype = None

libfibre_start_rx = lib.libfibre_start_rx
libfibre_start_rx.argtypes = [c_void_p, c_char_p, c_size_t, OnRxCompletedSignature, c_void_p]
libfibre_start_rx.restype = None

libfibre_cancel_rx = lib.libfibre_cancel_rx
libfibre_cancel_rx.argtypes = [c_void_p]
libfibre_cancel_rx.restype = None


# libfibre wrapper ------------------------------------------------------------#

class ObjectLostError(Exception):
    def __init__(self):
        super(Exception, self).__init__("the object disappeared")

def _get_exception(status):
    if status == kFibreOk:
        return None
    elif status == kFibreCancelled:
        return asyncio.CancelledError()
    elif status == kFibreClosed:
        return EOFError()
    elif status == kFibreInvalidArgument:
        return ArgumentError()
    elif status == kFibreInternalError:
        return Exception("internal libfibre error")
    elif status == kFibreProtocolError:
        return Exception("peer misbehaving")
    elif status == kFibreHostUnreachable:
        return ObjectLostError()
    else:
        return Exception("unknown libfibre error {}".format(status))

class StructCodec():
    """
    Generic serializer/deserializer based on struct pack
    """
    def __init__(self, struct_format, target_type):
        self._struct_format = struct_format
        self._target_type = target_type
    def get_length(self):
        return struct.calcsize(self._struct_format)
    def serialize(self, libfibre, value):
        if self._target_type == int and isinstance(value, enum.Enum):
            value = value.value
        value = self._target_type(value)
        return struct.pack(self._struct_format, value)
    def deserialize(self, libfibre, buffer):
        value = struct.unpack(self._struct_format, buffer)
        value = value[0] if len(value) == 1 else value
        return self._target_type(value)

class ObjectPtrCodec():
    """
    Serializer/deserializer for an object reference

    libfibre transcodes object references internally from/to something that can
    be sent over the wire and understood by the remote instance.
    """
    def get_length(self):
        return struct.calcsize("P")
    def serialize(self, libfibre, value):
        if value is None:
            return struct.pack("P", 0)
        elif isinstance(value, RemoteObject):
            assert(value._obj_handle) # Cannot serialize reference to a lost object
            return struct.pack("P", value._obj_handle)
        else:
            raise TypeError("Expected value of type RemoteObject or None but got '{}'. An example for a RemoteObject is this expression: mdrv0.axis0.controller._input_pos_property".format(type(value).__name__))
    def deserialize(self, libfibre, buffer):
        handle = struct.unpack("P", buffer)[0]

        # TODO: this is broken: A function can return an object ref before it is known to PyFibre.
        return None if handle == 0 else libfibre._objects[handle]


codecs = {
    'int8': StructCodec("<b", int),
    'uint8': StructCodec("<B", int),
    'int16': StructCodec("<h", int),
    'uint16': StructCodec("<H", int),
    'int32': StructCodec("<i", int),
    'uint32': StructCodec("<I", int),
    'int64': StructCodec("<q", int),
    'uint64': StructCodec("<Q", int),
    'bool': StructCodec("<?", bool),
    'float': StructCodec("<f", float),
    'object_ref': ObjectPtrCodec()
}

def decode_arg_list(arg_names, codec_names):
    for i in count(0):
        if arg_names[i] is None or codec_names[i] is None:
            break
        arg_name = arg_names[i].decode('utf-8')
        codec_name = codec_names[i].decode('utf-8')
        if not codec_name in codecs:
            raise Exception("unsupported codec {}".format(codec_name))
        yield arg_name, codec_name, codecs[codec_name]

def insert_with_new_id(dictionary, val):
    key = next(x for x in count(1) if x not in set(dictionary.keys()))
    dictionary[key] = val
    return key

# Runs a function on a foreign event loop and blocks until the function is done.
def run_coroutine_threadsafe(loop, func):
    future = concurrent.futures.Future()
    async def func_async():
        try:
            result = func()
            if hasattr(result, '__await__'):
                result = await result
            future.set_result(result)
        except Exception as ex:
            future.set_exception(ex)
    loop.call_soon_threadsafe(asyncio.ensure_future, func_async())
    return future.result()

class TxStream():
    """Python wrapper for libfibre's LibFibreTxStream interface"""

    def __init__(self, libfibre, tx_stream_handle):
        self._libfibre = libfibre
        self._tx_stream_handle = tx_stream_handle
        self._future = None
        self._tx_buf = None
        self._c_on_tx_completed = OnTxCompletedSignature(self._on_tx_completed)
        self.is_closed = False

    def _on_tx_completed(self, ctx, tx_stream, status, tx_end):
        tx_start = cast(self._tx_buf, c_void_p).value

        n_written = tx_end - tx_start
        assert(n_written <= len(self._tx_buf))
        future = self._future
        self._future = None
        self._tx_buf = None
        
        if status == kFibreClosed:
            self.is_closed = True
        
        if status == kFibreOk or status == kFibreClosed:
            future.set_result(n_written)
        else:
            future.set_exception(_get_exception(status))

    def write(self, data):
        """
        Writes the provided data to the stream. Not all bytes are guaranteed to
        be written. The caller should check the return value to determine the
        actual number of bytes written.

        If a non-empty buffer is provided, this function will either write at
        least one byte to the output, set is_closed to True or throw an
        Exception (through the future).

        Currently only one write call may be active at a time (this may change
        in the future).

        Returns: A future that completes with the number of bytes actually
        written or an Exception.
        """
        assert(self._future is None)
        self._future = future = self._libfibre.loop.create_future()
        self._tx_buf = data # Retain a reference to the buffer to prevent it from being garbage collected

        libfibre_start_tx(self._tx_stream_handle,
            cast(self._tx_buf, c_char_p), len(self._tx_buf),
            self._c_on_tx_completed, None)

        return future
        
    async def write_all(self, data):
        """
        Writes all of the provided data to the stream or completes with an
        Exception.

        If an empty buffer is provided, the underlying stream's write function
        is still called at least once.

        Returns: A future that either completes with an empty result or with
        an Exception.
        """

        while True:
            n_written = await self.write(data)
            data = data[n_written:]
            if len(data) == 0:
                break
            elif self.is_closed:
                raise EOFError("the TX stream was closed but there are still {} bytes left to send".format(len(data)))
            assert(n_written > 0) # Ensure progress

class RxStream():
    """Python wrapper for libfibre's LibFibreRxStream interface"""

    def __init__(self, libfibre, rx_stream_handle):
        self._libfibre = libfibre
        self._rx_stream_handle = rx_stream_handle
        self._future = None
        self._rx_buf = None
        self._c_on_rx_completed = OnRxCompletedSignature(self._on_rx_completed)
        self.is_closed = False

    def _on_rx_completed(self, ctx, rx_stream, status, rx_end):
        rx_start = cast(self._rx_buf, c_void_p).value

        n_read = rx_end - rx_start
        assert(n_read <= len(self._rx_buf))
        data = self._rx_buf[:n_read]
        future = self._future
        self._future = None
        self._rx_buf = None
        
        if status == kFibreClosed:
            self.is_closed = True
        
        if status == kFibreOk or status == kFibreClosed:
            future.set_result(data)
        else:
            future.set_exception(_get_exception(status))

    def read(self, n_read):
        """
        Reads up to the specified number of bytes from the stream.

        If more than zero bytes are requested, this function will either read at
        least one byte, set is_closed to True or throw an Exception (through the
        future).

        Currently only one write call may be active at a time (this may change
        in the future).

        Returns: A future that either completes with a buffer containing the
        bytes that were read or completes with an Exception.
        """
        assert(self._future is None)
        self._future = future = self._libfibre.loop.create_future()
        self._rx_buf = bytes(n_read)

        libfibre_start_rx(self._rx_stream_handle,
            cast(self._rx_buf, c_char_p), len(self._rx_buf),
            self._c_on_rx_completed, None)

        return future
        
    async def read_all(self, n_read):
        """
        Reads the specified number of bytes from the stream or throws an
        Exception.

        If zero bytes are requested, the underlying stream's read function
        is still called at least once.

        Returns: A future that either completes with a buffer of size n_read or
        an Exception.
        """

        data = bytes()
        while True:
            chunk = await self.read(n_read - len(data))
            data += chunk
            if n_read == len(data):
                break
            elif self.is_closed:
                raise EOFError()
            assert(len(chunk) > 0) # Ensure progress
        return data


class Call(object):
    """
    This call behaves as you would expect an async generator to behave. This is
    used to provide compatibility down to Python 3.5.
    """
    def __init__(self, func):
        self._func = func
        self._call_handle = c_void_p(0)
        self._is_started = False
        self._should_close = False
        self._is_closed = False
        self._tx_buf = None

    def __aiter__(self):
        return self

    async def asend(self, val):
        assert(self._is_started == (not val is None))
        if not val is None:
            self._tx_buf, self._rx_len, self._should_close = val
        return await self.__anext__()

    async def __anext__(self):
        if not self._is_started:
            self._is_started = True
            return None # This immitates the weird starting behavior of Python 3.6+ async generators iterators

        if self._is_closed:
            raise StopAsyncIteration

        tx_end = c_void_p(0)
        rx_end = c_void_p(0)

        rx_buf = bytes(self._rx_len)

        call_id = insert_with_new_id(self._func._libfibre._calls, self)

        status = libfibre_call(self._func._func_handle, byref(self._call_handle),
                kFibreClosed if self._should_close else kFibreOk,
                cast(self._tx_buf, c_char_p), len(self._tx_buf),
                cast(rx_buf, c_char_p), len(rx_buf),
                byref(tx_end), byref(rx_end), self._func._libfibre.c_on_call_completed, call_id)

        if status == kFibreBusy:
            self.ag_await = self._func._libfibre.loop.create_future()
            status, tx_end, rx_end = await self.ag_await
            self.ag_await = None

        if status != kFibreOk and status != kFibreClosed:
            raise _get_exception(status)

        n_written = tx_end - cast(self._tx_buf, c_void_p).value
        self._tx_buf = self._tx_buf[n_written:]
        n_read = rx_end - cast(rx_buf, c_void_p).value
        rx_buf = rx_buf[:n_read]

        if status != kFibreOk:
            self._is_closed = True
        return self._tx_buf, rx_buf, self._is_closed

    async def cancel():
        # TODO: this doesn't follow the official Python async generator protocol. Should implement aclose() instead.
        status = libfibre_call(self._func._func_handle, byref(self._call_handle), kFibreOk,
                0, 0, 0, 0, 0, 0, self._func._libfibre.c_on_call_completed, call_id)

    #async def aclose(self):
    #    assert(self._is_started and not self._is_closed)
    #    return self._tx_buf, rx_buf, self._is_closed


class RemoteFunction(object):
    """
    Represents a callable function that maps to a function call on a remote object.
    """
    def __init__(self, libfibre, func_handle, inputs, outputs):
        self._libfibre = libfibre
        self._func_handle = func_handle
        self._inputs = inputs
        self._outputs = outputs
        self._rx_size = sum(codec.get_length() for _, _, codec in self._outputs)

    async def async_call(self, args, cancellation_token):
        #print("making call on " + hex(args[0]._obj_handle))
        tx_buf = bytes()
        for i, arg in enumerate(self._inputs):
            tx_buf += arg[2].serialize(self._libfibre, args[i])
        rx_buf = bytes()

        agen = Call(self)
        
        if not cancellation_token is None:
            cancellation_token.add_done_callback(agen.cancel)

        try:
            assert(await agen.asend(None) is None)

            is_closed = False
            while not is_closed:
                tx_buf, rx_chunk, is_closed = await agen.asend((tx_buf, self._rx_size - len(rx_buf), True))
                rx_buf += rx_chunk

        finally:
            if not cancellation_token is None:
                cancellation_token.remove_done_callback(agen.cancel)

        assert(len(rx_buf) == self._rx_size)

        outputs = []
        for arg in self._outputs:
            arg_length = arg[2].get_length()
            outputs.append(arg[2].deserialize(self._libfibre, rx_buf[:arg_length]))
            rx_buf = rx_buf[arg_length:]

        if len(outputs) == 0:
            return
        elif len(outputs) == 1:
            return outputs[0]
        else:
            return tuple(outputs)

    def __call__(self, *args, cancellation_token = None):
        """
        Starts invoking the remote function. The first argument is usually a
        remote object.
        If this function is called from the Fibre thread then it is nonblocking
        and returns an asyncio.Future. If it is called from another thread then
        it blocks until the function completes and returns the result(s) of the 
        invokation.
        """

        if threading.current_thread() != libfibre_thread:
            return run_coroutine_threadsafe(self._libfibre.loop, lambda: self.__call__(*args))

        if (len(self._inputs) != len(args)):
            raise TypeError("expected {} arguments but have {}".format(len(self._inputs), len(args)))

        coro = self.async_call(args, cancellation_token)
        return asyncio.ensure_future(coro, loop=self._libfibre.loop)

    def __get__(self, instance, owner):
        return MethodType(self, instance) if instance else self

    def _dump(self, name):
        print_arglist = lambda arglist: ", ".join("{}: {}".format(arg_name, codec_name) for arg_name, codec_name, codec in arglist)
        return "{}({}){}".format(name,
            print_arglist(self._inputs),
            "" if len(self._outputs) == 0 else
            " -> " + print_arglist(self._outputs) if len(self._outputs) == 1 else
            " -> (" + print_arglist(self._outputs) + ")")

class RemoteAttribute(object):
    def __init__(self, libfibre, attr_handle, intf_handle, intf_name, magic_getter, magic_setter):
        self._libfibre = libfibre
        self._attr_handle = attr_handle
        self._intf_handle = intf_handle
        self._intf_name = intf_name
        self._magic_getter = magic_getter
        self._magic_setter = magic_setter

    def _get_obj(self, instance):
        assert(not instance._obj_handle is None)

        obj_handle = c_void_p(0)
        status = libfibre_get_attribute(instance._obj_handle, self._attr_handle, byref(obj_handle))
        if status != kFibreOk:
            raise _get_exception(status)
        
        obj = self._libfibre._load_py_obj(obj_handle.value, self._intf_handle)
        if obj in instance._children:
            self._libfibre._release_py_obj(obj_handle.value)
        else:
            # the object will be released when the parent is released
            instance._children.add(obj)

        return obj

    def __get__(self, instance, owner):
        if not instance:
            return self

        if self._magic_getter:
            if threading.current_thread() == libfibre_thread:
                # read() behaves asynchronously when run on the fibre thread
                # which means it returns an awaitable which _must_ be awaited
                # (otherwise it's a bug). However hasattr(...) internally calls
                # __get__ and does not await the result. Thus the safest thing
                # is to just disallow __get__ from run as an async method.
                raise Exception("Cannot use magic getter on Fibre thread. Use _[prop_name]_propery.read() instead.")
            return self._get_obj(instance).read()
        else:
            return self._get_obj(instance)

    def __set__(self, instance, val):
        if self._magic_setter:
            return self._get_obj(instance).exchange(val)
        else:
            raise Exception("this attribute cannot be written to")

class EmptyInterface():
    def __str__(self):
        return "[lost object]"
    def __repr__(self):
        return self.__str__()

class RemoteObject(object):
    """
    Base class for interfaces of remote objects.
    """
    __sealed__ = False

    def __init__(self, libfibre, obj_handle):
        self.__class__._refcount += 1
        self._refcount = 0
        self._children = set()

        self._libfibre = libfibre
        self._obj_handle = obj_handle
        self._on_lost = concurrent.futures.Future() # TODO: maybe we can do this with conc

        # Ensure that assignments to undefined attributes raise an exception
        self.__sealed__ = True

    def __setattr__(self, key, value):
        if self.__sealed__ and not key in dir(self) and not hasattr(self, key):
            raise AttributeError("Attribute {} not found".format(key))
        object.__setattr__(self, key, value)

    #def __del__(self):
    #    print("unref")
    #    libfibre_unref_obj(self._obj_handle)

    def _dump(self, indent, depth):
        if self._obj_handle is None:
            return "[object lost]"

        try:
            if depth <= 0:
                return "..."
            lines = []
            for key in dir(self.__class__):
                if key.startswith('_'):
                    continue
                class_member = getattr(self.__class__, key)
                if isinstance(class_member, RemoteFunction):
                    lines.append(indent + class_member._dump(key))
                elif isinstance(class_member, RemoteAttribute):
                    val = getattr(self, key)
                    if isinstance(val, RemoteObject) and not class_member._magic_getter:
                        lines.append(indent + key + (": " if depth == 1 else ":\n") + val._dump(indent + "  ", depth - 1))
                    else:
                        if isinstance(val, RemoteObject) and class_member._magic_getter:
                            val_str = get_user_name(val)
                        else:
                            val_str = str(val)
                        property_type = str(class_member._get_obj(self).__class__.read._outputs[0][1])
                        lines.append(indent + key + ": " + val_str + " (" + property_type + ")")
                else:
                    lines.append(indent + key + ": " + str(type(val)))
        except:
            return "[failed to dump object]"

        return "\n".join(lines)

    def __str__(self):
        return self._dump("", depth=2)

    def __repr__(self):
        return self.__str__()

    def _destroy(self):
        libfibre = self._libfibre
        on_lost = self._on_lost
        children = self._children

        self._libfibre = None
        self._obj_handle = None
        self._on_lost = None
        self._children = set()

        for child in children:
            libfibre._release_py_obj(child._obj_handle)

        self.__class__._refcount -= 1
        if self.__class__._refcount == 0:
            libfibre.interfaces.pop(self.__class__._handle)

        self.__class__ = EmptyInterface # ensure that this object has no more attributes
        on_lost.set_result(True)


class LibFibre():
    def __init__(self, loop):
        self.loop = loop

        # We must keep a reference to these function objects so they don't get
        # garbage collected.
        self.c_post = PostSignature(self._post)
        self.c_register_event = RegisterEventSignature(self._register_event)
        self.c_deregister_event = DeregisterEventSignature(self._deregister_event)
        self.c_call_later = CallLaterSignature(self._call_later)
        self.c_cancel_timer = CancelTimerSignature(self._cancel_timer)
        self.c_on_found_object = OnFoundObjectSignature(self._on_found_object)
        self.c_on_lost_object = OnLostObjectSignature(self._on_lost_object)
        self.c_on_discovery_stopped = OnStoppedSignature(self._on_discovery_stopped)
        self.c_on_attribute_added = OnAttributeAddedSignature(self._on_attribute_added)
        self.c_on_attribute_removed = OnAttributeRemovedSignature(self._on_attribute_removed)
        self.c_on_function_added = OnFunctionAddedSignature(self._on_function_added)
        self.c_on_function_removed = OnFunctionRemovedSignature(self._on_function_removed)
        self.c_on_call_completed = OnCallCompletedSignature(self._on_call_completed)
        
        self.timer_map = {}
        self.eventfd_map = {}
        self.interfaces = {} # key: libfibre handle, value: python class
        self.discovery_processes = {} # key: ID, value: python dict
        self._objects = {} # key: libfibre handle, value: python class
        self._calls = {} # key: libfibre handle, value: Call object

        event_loop = LibFibreEventLoop()
        event_loop.post = self.c_post
        event_loop.register_event = self.c_register_event
        event_loop.deregister_event = self.c_deregister_event
        event_loop.call_later = self.c_call_later
        event_loop.cancel_timer = self.c_cancel_timer

        self.ctx = c_void_p(libfibre_open(event_loop))
        assert(self.ctx)

    def _post(self, callback, ctx):
        self.loop.call_soon_threadsafe(callback, ctx)
        return 0

    def _register_event(self, event_fd, events, callback, ctx):
        self.eventfd_map[event_fd] = events
        if (events & 1):
            self.loop.add_reader(event_fd, lambda x: callback(x, 1), ctx)
        if (events & 4):
            self.loop.add_writer(event_fd, lambda x: callback(x, 4), ctx)
        if (events & 0xfffffffa):
            raise Exception("unsupported event mask " + str(events))
        return 0

    def _deregister_event(self, event_fd):
        events = self.eventfd_map.pop(event_fd)
        if (events & 1):
            self.loop.remove_reader(event_fd)
        if (events & 4):
            self.loop.remove_writer(event_fd)
        return 0

    def _call_later(self, delay, callback, ctx):
        def cb():
            self.timer_map.pop(timer_id)
            callback(ctx)
        timer_id = insert_with_new_id(self.timer_map, self.loop.call_later(delay, cb))
        return timer_id

    def _cancel_timer(self, timer_id):
        self.timer_map.pop(timer_id).cancel()
        return 0

    def _load_py_intf(self, name, intf_handle):
        """
        Creates a new python type for the specified libfibre interface handle or
        returns the existing python type if one was already create before.

        Behind the scenes the python type will react to future events coming
        from libfibre, such as functions/attributes being added/removed.
        """
        if intf_handle in self.interfaces:
            return self.interfaces[intf_handle]
        else:
            if name is None:
                name = "anonymous_interface_" + str(intf_handle)
            py_intf = self.interfaces[intf_handle] = type(name, (RemoteObject,), {'_handle': intf_handle, '_refcount': 0})
            #exit(1)
            libfibre_subscribe_to_interface(intf_handle, self.c_on_attribute_added, self.c_on_attribute_removed, self.c_on_function_added, self.c_on_function_removed, intf_handle)
            return py_intf

    def _load_py_obj(self, obj_handle, intf_handle):
        if not obj_handle in self._objects:
            name = None # TODO: load from libfibre
            py_intf = self._load_py_intf(name, intf_handle)
            py_obj = py_intf(self, obj_handle)
            self._objects[obj_handle] = py_obj
        else:
            py_obj = self._objects[obj_handle]

        # Note: this refcount does not count the python references to the object
        # but rather mirrors the libfibre-internal refcount of the object. This
        # is so that we can destroy the Python object when libfibre releases it.
        py_obj._refcount += 1
        return py_obj

    def _release_py_obj(self, obj_handle):
        py_obj = self._objects[obj_handle]
        py_obj._refcount -= 1
        if py_obj._refcount <= 0:
            self._objects.pop(obj_handle)
            py_obj._destroy()

    def _on_found_object(self, ctx, obj, intf):
        py_obj = self._load_py_obj(obj, intf)
        discovery = self.discovery_processes[ctx]
        

        # TODO: this is a hack because ObjectPtrCodec is broken
        def load(subobj):
            for key in dir(subobj):
                if not key.startswith('_'):
                    attr = getattr(subobj.__class__, key)
                    if isinstance(attr, RemoteAttribute):
                        load(attr._get_obj(subobj))

        load(py_obj)

        discovery._unannounced.append(py_obj)
        old_future = discovery._future
        discovery._future = self.loop.create_future()
        old_future.set_result(None)
    
    def _on_lost_object(self, ctx, obj):
        assert(obj)
        self._release_py_obj(obj)
    
    def _on_discovery_stopped(self, ctx, result):
        print("discovery stopped")

    def _on_attribute_added(self, ctx, attr, name, name_length, subintf, subintf_name, subintf_name_length):
        name = string_at(name, name_length).decode('utf-8')
        subintf_name = None if subintf_name is None else string_at(subintf_name, subintf_name_length).decode('utf-8')
        intf = self.interfaces[ctx]

        magic_getter = not subintf_name is None and subintf_name.startswith("fibre.Property<") and subintf_name.endswith(">")
        magic_setter = not subintf_name is None and subintf_name.startswith("fibre.Property<readwrite ") and subintf_name.endswith(">")

        setattr(intf, name, RemoteAttribute(self, attr, subintf, subintf_name, magic_getter, magic_setter))
        if magic_getter or magic_setter:
            setattr(intf, "_" + name + "_property", RemoteAttribute(self, attr, subintf, subintf_name, False, False))

    def _on_attribute_removed(self, ctx, attr):
        print("attribute removed") # TODO

    def _on_function_added(self, ctx, func, name, name_length, input_names, input_codecs, output_names, output_codecs):
        name = string_at(name, name_length).decode('utf-8')
        inputs = list(decode_arg_list(input_names, input_codecs))
        outputs = list(decode_arg_list(output_names, output_codecs))
        intf = self.interfaces[ctx]
        setattr(intf, name, RemoteFunction(self, func, inputs, outputs))

    def _on_function_removed(self, ctx, func):
        print("function removed") # TODO

    def _on_call_completed(self, ctx, status, tx_end, rx_end, tx_buf, tx_len, rx_buf, rx_len):
        call = self._calls.pop(ctx)

        call.ag_await.set_result((status, tx_end, rx_end))

        return kFibreBusy

class Discovery():
    """
    All public members of this class are thread-safe.
    """

    def __init__(self, domain):
        self._domain = domain
        self._id = 0
        self._discovery_handle = c_void_p(0)
        self._unannounced = []
        self._future = domain._libfibre.loop.create_future()

    async def _next(self):
        if len(self._unannounced) == 0:
            await self._future
        return self._unannounced.pop(0)

    def _stop(self):
        self._domain._libfibre.discovery_processes.pop(self._id)
        libfibre_stop_discovery(self._discovery_handle)
        self._future.set_exception(asyncio.CancelledError())

    def stop(self):
        if threading.current_thread() == libfibre_thread:
            self._stop()
        else:
            run_coroutine_threadsafe(self._domain._libfibre.loop, self._stop)

class _Domain():
    """
    All public members of this class are thread-safe.
    """

    def __init__(self, libfibre, handle):
        self._libfibre = libfibre
        self._domain_handle = handle

    def _close(self):
        libfibre_close_domain(self._domain_handle)
        self._domain_handle = None
        #decrement_lib_refcount()

    def _start_discovery(self):
        discovery = Discovery(self)
        discovery._id = insert_with_new_id(self._libfibre.discovery_processes, discovery)
        libfibre_start_discovery(self._domain_handle, byref(discovery._discovery_handle), self._libfibre.c_on_found_object, self._libfibre.c_on_lost_object, self._libfibre.c_on_discovery_stopped, discovery._id)
        return discovery

    async def _discover_one(self):
        discovery = self._start_discovery()
        obj = await discovery._next()
        discovery._stop()
        return obj

    def discover_one(self):
        """
        Blocks until exactly one object is discovered.
        """
        return run_coroutine_threadsafe(self._libfibre.loop, self._discover_one)

    def run_discovery(self, callback):
        """
        Invokes `callback` for every object that is discovered. The callback is
        invoked on the libfibre thread and can be an asynchronous function.
        Returns a `Discovery` object on which `stop()` can be called to
        terminate the discovery.
        """
        discovery = run_coroutine_threadsafe(self._libfibre.loop, self._start_discovery)
        async def loop():
            while True:
                obj = await discovery._next()
                await callback(obj)
        self._libfibre.loop.call_soon_threadsafe(lambda: asyncio.ensure_future(loop()))
        return discovery


class Domain():
    def __init__(self, path):
        increment_lib_refcount()
        self._opened_domain = run_coroutine_threadsafe(libfibre.loop, lambda: Domain._open(path))
        
    def _open(path):
        assert(libfibre_thread == threading.current_thread())
        buf = path.encode('ascii')
        domain_handle = libfibre_open_domain(libfibre.ctx, buf, len(buf))
        return _Domain(libfibre, domain_handle)

    def __enter__(self):
        return self._opened_domain

    def __exit__(self, type, value, traceback):
        run_coroutine_threadsafe(self._opened_domain._libfibre.loop, self._opened_domain._close)
        self._opened_domain = None
        decrement_lib_refcount()

libfibre = None

def _run_event_loop():
    global libfibre
    global terminate_libfibre

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    terminate_libfibre = loop.create_future()
    libfibre = LibFibre(loop)

    libfibre.loop.run_until_complete(terminate_libfibre)

    libfibre_close(libfibre.ctx)

    # Detach all objects that still exist
    # TODO: the proper way would be either of these
    #  - provide a libfibre function to destroy an object on-demand which we'd
    #    call before libfibre_close().
    #  - have libfibre_close() report the destruction of all objects

    while len(libfibre._objects):
        libfibre._objects.pop(list(libfibre._objects.keys())[0])._destroy()
    assert(len(libfibre.interfaces) == 0)

    libfibre = None


lock = threading.Lock()
libfibre_refcount = 0
libfibre_thread = None

def increment_lib_refcount():
    global libfibre_refcount
    global libfibre_thread

    with lock:
        libfibre_refcount += 1
        #print("inc refcount to {}".format(libfibre_refcount))

        if libfibre_refcount == 1:
            libfibre_thread = threading.Thread(target = _run_event_loop)
            libfibre_thread.start()

        while libfibre is None:
            time.sleep(0.1)

def decrement_lib_refcount():
    global libfibre_refcount
    global libfibre_thread

    with lock:
        #print("dec refcount from {}".format(libfibre_refcount))
        libfibre_refcount -= 1

        if libfibre_refcount == 0:
            libfibre.loop.call_soon_threadsafe(lambda: terminate_libfibre.set_result(True))

            # It's unlikely that releasing fibre from a fibre callback is ok. If
            # there is a valid scenario for this then we can remove the assert.
            assert(libfibre_thread != threading.current_thread())

            libfibre_thread.join()
            libfibre_thread = None

def get_user_name(obj):
    """
    Can be overridden by the application to return the user-facing name of an
    object.
    """
    return "[anonymous object]"



from setuptools import setup, find_packages
import io
import re
import os

here = os.path.abspath(os.path.dirname(__file__))

def read(fname):
    with io.open(os.path.join(here, fname), encoding='utf-8') as f:
        return f.read()

def read_version():
    ver_file = os.path.join(here, "mwdrive", "version.py")
    text = read(os.path.relpath(ver_file, here))
    m = re.search(r'__version__\s*=\s*[\'"]([^\'"]+)[\'"]', text)
    if m:
        return m.group(1)
    raise RuntimeError("Unable to find version string.")

setup(
    name="mwdrive",  # <-- 改为你要上传到 PyPI 的包名
    version=read_version(),
    description="Motor Driver Control Library for products from CyberBeast",
    long_description=read("README.md") if os.path.exists(os.path.join(here, "README.md")) else "",
    long_description_content_type="text/markdown",
    author="CyberBeast",
    packages=find_packages(),  # 会在项目根找到 mwdrive/
    include_package_data=True,
    scripts=[
        "scripts/mtool",
        "scripts/mtool.bat",
    ],
    install_requires=[
        # 在这里列出运行时依赖，例如 "pyusb>=1.2.1"
    ],
    # 如果希望安装后生成命令行脚本，可在此处添加 entry_points
    # entry_points={
    #     "console_scripts": [
    #         "mtool=mwdrive.mtool:main",
    #     ],
    # },
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
)




import sys
import time
import threading
import platform
import subprocess
import os

try:
    if platform.system() == 'Windows':
        import win32console
        # TODO: we should win32console anyway so we could just omit colorama
        import colorama
        colorama.init()
except ImportError:
    print("Could not init terminal features.")
    sys.stdout.flush()
    pass


## Log utils ##

# TODO: use "logging" from Python standard library instead of this
class Logger():
    """
    Logs messages to stdout
    """

    COLOR_DEFAULT = 0
    COLOR_GREEN = 1
    COLOR_CYAN = 2
    COLOR_YELLOW = 3
    COLOR_RED = 4

    _VT100Colors = {
        COLOR_GREEN: '\x1b[92;1m',
        COLOR_CYAN: '\x1b[96;1m',
        COLOR_YELLOW: '\x1b[93;1m',
        COLOR_RED: '\x1b[91;1m',
        COLOR_DEFAULT: '\x1b[0m'
    }

    _Win32Colors = {
        COLOR_GREEN: 0x0A,
        COLOR_CYAN: 0x0B,
        COLOR_YELLOW: 0x0E,
        COLOR_RED: 0x0C,
        COLOR_DEFAULT: 0x07
    }

    def __init__(self, verbose=True):
        self._prefix = ''
        self._skip_bottom_line = False # If true, messages are printed one line above the cursor
        self._verbose = verbose
        self._print_lock = threading.Lock()
        if platform.system() == 'Windows':
            self._stdout_buf = win32console.GetStdHandle(win32console.STD_OUTPUT_HANDLE)

    def indent(self, prefix='  '):
        indented_logger = Logger()
        indented_logger._prefix = self._prefix + prefix
        return indented_logger

    def print_on_second_last_line(self, text, color):
        """
        Prints a text on the second last line.
        This can be used to print a message above the command
        prompt. If the command prompt spans multiple lines
        there will be glitches.
        If the printed text spans multiple lines there will also
        be glitches (though this could be fixed).
        """

        if platform.system() == 'Windows':
            # Windows <10 doesn't understand VT100 escape codes and the colorama
            # also doesn't support the specific escape codes we need so we use the
            # native Win32 API.
            info = self._stdout_buf.GetConsoleScreenBufferInfo()
            cursor_pos = info['CursorPosition']
            scroll_rect=win32console.PySMALL_RECTType(
                Left=0, Top=1,
                Right=info['Window'].Right,
                Bottom=cursor_pos.Y-1)
            scroll_dest = win32console.PyCOORDType(scroll_rect.Left, scroll_rect.Top-1)
            self._stdout_buf.ScrollConsoleScreenBuffer(
                scroll_rect, scroll_rect, scroll_dest, # clipping rect is same as scroll rect
                u' ', Logger._Win32Colors[color]) # fill with empty cells with the desired color attributes
            line_start = win32console.PyCOORDType(0, cursor_pos.Y-1)
            self._stdout_buf.WriteConsoleOutputCharacter(text, line_start)

        else:
            # Assume we're in a terminal that interprets VT100 escape codes.
            # TODO: test on macOS

            # Escape character sequence:
            #   ESC 7: store cursor position
            #   ESC 1A: move cursor up by one
            #   ESC 1S: scroll entire viewport by one
            #   ESC 1L: insert 1 line at cursor position
            #   (print text)
            #   ESC 8: restore old cursor position

            self._print_lock.acquire()
            sys.stdout.write('\x1b7\x1b[1A\x1b[1S\x1b[1L')
            sys.stdout.write(Logger._VT100Colors[color] + text + Logger._VT100Colors[Logger.COLOR_DEFAULT])
            sys.stdout.write('\x1b8')
            sys.stdout.flush()
            self._print_lock.release()

    def print_colored(self, text, color):
        if self._skip_bottom_line:
            self.print_on_second_last_line(text, color)
        else:
            # On Windows, colorama does the job of interpreting the VT100 escape sequences
            self._print_lock.acquire()
            sys.stdout.write(Logger._VT100Colors[color] + text + Logger._VT100Colors[Logger.COLOR_DEFAULT] + '\n')
            sys.stdout.flush()
            self._print_lock.release()

    def debug(self, text):
        if self._verbose:
            self.print_colored(self._prefix + text, Logger.COLOR_DEFAULT)
    def success(self, text):
        self.print_colored(self._prefix + text, Logger.COLOR_GREEN)
    def info(self, text):
        self.print_colored(self._prefix + text, Logger.COLOR_DEFAULT)
    def notify(self, text):
        self.print_colored(self._prefix + text, Logger.COLOR_CYAN)
    def warn(self, text):
        self.print_colored(self._prefix + text, Logger.COLOR_YELLOW)
    def error(self, text):
        # TODO: write to stderr
        self.print_colored(self._prefix + text, Logger.COLOR_RED)




import re
from typing import List, NamedTuple

class HwVersion(NamedTuple):
    """
    Represents a hardware version triplet.
    See also hw_version.hpp and hw_version.dart.
    """
    product_line: int
    version: int
    variant: int

    @staticmethod
    def from_string(arg: str):
        """
        Constructs a HwVersion from a string of the form "4.4.58".
        """
        return HwVersion(*(int(i) for i in re.match(r'^([0-9]+)\.([0-9]+).([0-9]+)$', arg).groups()))

    @staticmethod
    def from_tuple(arg: List):
        """
        Constructs a HwVersion from a tuple or list of the form (4, 4, 58).
        """
        assert len(arg) == 3
        assert all([type(a) == int for a in arg])
        return HwVersion(*arg)

    @staticmethod
    def from_json(json):
        return HwVersion.from_tuple(json)

    @property
    def display_name(self):
        """
        Returns a display name such as "ODrive Pro" or "unknown device"
        corresponding to this board version.
        See also hw_version.hpp.
        """
        if self.product_line == 3:
            return f"CyberBeast Leopard v3.{self.version}"
        elif self.product_line == 4:
            return {
                0: "CyberBeast Lion v4.0",
                1: "CyberBeast Lion v4.1",
                2: "CyberBeast Lion v4.2",
                3: "CyberBeast Lion v4.3",
                4: "CyberBeast Lion",
            }.get(self.version, "unknown CyberBeast Lion v4")
        elif self.product_line == 5:
            return {
                0: "CyberBeast Leopard v5.0",
                1: "CyberBeast Leopard v5.1",
                2: "CyberBeast Leopard v5.2",
            }.get(self.version, "unknown CyberBeast Leopard v5")
        elif self.product_line == 6:
            return {
                0: "CyberBeast Lion v6.0",
                1: "CyberBeast Lion v6.1",
            }.get(self.version, "unknown CyberBeast Lion v6")
        elif self.product_line == 8:
            return {
                1: "CyberBeast Tiger v8.1",
            }.get(self.version, "unknown CyberBeast Tiger v8")
        else:
            return "unknown device"

    def to_json(self):
        return [self.product_line, self.version, self.variant]




import asyncio
import os
import sys
import platform
import threading
import fibre
import mwdrive
from . import config, enums, rich_text, utils

def print_banner():
    print('官方网站(Website): https://www.cyberbeast.cn')
    print('用户手册(User Manual): https://cyberbeast.feishu.cn/docx/N3SMd4QyRobzHkx3wP3cT1qXnpf')
    print('协议手册(Protocol Manual): https://cyberbeast.feishu.cn/docx/BPnQd8reEotLWVxqHFNc9qYZnKh')

    print()
    print('请连接您的驱动器')
    print('输入help()获取操作帮助, exit退出')
    print('Please connect your MWDrive')
    print('You can also type help() or exit')

def print_help(args, have_devices):
    print('')
    if have_devices:
        print('连接到 {} 并打开电源, 然后会出现下述提示:'.format(args.path))
        print('  "Connected to MWDrive [serial number] as mdrv0"')
        print('')
        print('连接成功后, 输入"mdrv0."然后按<TAB>键可以查看所有可用参数/方法')
        print('')
        print('Connect your MWDrive to {} and power it up.'.format(args.path))
        print('After that, the following message should appear:')
        print('  "Connected to MWDrive [serial number] as mdrv0"')
        print('')
        print('Once the MWDrive is connected, type "mdrv0." and press <tab>')
        print('This will present you with all the properties that you can reference')
    else:
        print('输入"mdrv0."然后按<TAB>键可以查看所有可用参数/方法')
        print('Type "mdrv0." and press <tab>')
        print('This will present you with all the properties that you can reference')
    print('')
    print('例如: "mdrv0.axis0.encoder.pos_estimate"可以查看轴0的编码器位置')
    print('"mdrv0.axis0.controller.input_pos = 0.5" 可以让轴0转动到0.5圈位置')
    print('请注意, 力矩的单位是Nm, 速度的单位是圈/秒(电机转子侧), 位置的单位是圈(电机转子侧)')
    print('')
    print('For example: "mdrv0.axis0.encoder.pos_estimate"')
    print('will print the current encoder position on axis 0')
    print('and "mdrv0.axis0.controller.input_pos = 0.5"')
    print('will send axis 0 to 0.5 turns')
    print('')


interactive_variables = {}
discovered_devices = []

def benchmark(mdrv):
    import time

    async def measure_async():
        start = time.monotonic()
        futures = [mdrv.vbus_voltage for i in range(1000)]
#        data = [await f for f in futures]
#        print("took " + str(time.monotonic() - start) + " seconds. Average is " + str(sum(data) / len(data)))

    fibre.libfibre.libfibre.loop.call_soon_threadsafe(lambda: asyncio.ensure_future(measure_async()))

class ShellVariables():
    mwdrive = mwdrive
    devices = []
    config = config.MachineConfig()

    @staticmethod
    def apply():
        ShellVariables.config.apply(ShellVariables.devices)

    @staticmethod
    def status():
        rich_text.print_rich_text(ShellVariables.config.format_status(ShellVariables.devices))

    @staticmethod
    def calibrate():
        ShellVariables.config.calibrate(ShellVariables.devices)

def _import_from(source):
    return {
        k: getattr(source, k)
        for k in dir(source)
        if not k.startswith("_")
    }

def _wrap_async(func):
    def wrapper(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))
    return wrapper

def launch_shell(args, logger):
    """
    Launches an interactive python or IPython command line
    interface.
    As MDrives are connected they are made available as
    "mdrv0", "mdrv1", ...
    """

    interactive_variables = {
        'start_liveplotter': utils.start_liveplotter,
        'dump_errors': utils.dump_errors,
        'benchmark': benchmark,
        'oscilloscope_dump': utils.oscilloscope_dump,
        'dump_interrupts': utils.dump_interrupts,
        'dump_threads': utils.dump_threads,
        'dump_dma': utils.dump_dma,
        'dump_timing': utils.dump_timing,
        'ram_osci_config': _wrap_async(utils.ram_osci_config),
        'ram_osci_trigger': utils.ram_osci_trigger,
        'ram_osci_download': _wrap_async(utils.ram_osci_download),
        'ram_osci_run': _wrap_async(utils.ram_osci_run),
        'BulkCapture': utils.BulkCapture,
        'step_and_plot': utils.step_and_plot,
        'calculate_thermistor_coeffs': utils.calculate_thermistor_coeffs,
        'set_motor_thermistor_coeffs': utils.set_motor_thermistor_coeffs,
    }

    # Import a bunch of variables and functions from various sources
    interactive_variables.update(_import_from(ShellVariables))
    interactive_variables.update(_import_from(enums))
    interactive_variables.update(_import_from(config))

    private_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'mdrive_private')
    if os.path.isfile(os.path.join(private_path, '__init__.py')):
        print("loading private plugins...")
        sys.path.insert(0, private_path)
        import mdrive_private
        mdrive_private.load_mtool_plugins(interactive_variables)

    async def mount(obj):
        serial_number_str = await utils.get_serial_number_str(obj)
        if not args.serial_number is None and serial_number_str != args.serial_number:
            return None
        if hasattr(obj, '_otp_valid_property') and (not await obj._otp_valid_property.read()):
            logger.warn('Device {}: Not a genuine MWDrive! Some features may not work as expected.'.format(serial_number_str))
            return ('device ' + serial_number_str, serial_number_str, 'dev')
        await utils.attach_metadata(obj)
        fw_version_str = '???' if obj._fw_version is None else 'v{}.{}.{}'.format(*obj._fw_version)
        return (f"{obj._board.display_name if not obj._board is None else 'device'} {serial_number_str} (firmware {fw_version_str})", serial_number_str, "mdrv")

    fibre.launch_shell(args, mount,
                       interactive_variables,
                       ShellVariables.devices,
                       print_banner, print_help,
                       logger)
